'''Handle calls and json convertions.'''
import json
from typing import Dict
from fastapi import status
from karp import auth


def get_json(
    client,
    path: str,
    expected_status_code: int = status.HTTP_200_OK,
    **kwargs
):
    '''Call the get on the client with the given path.

    Assert that the status_code is ok.
    Then load the json response.

    Arguments:
        client {[type]} -- [description]
        path {[type]} -- [description]

    Returns:
        [type] -- [description]
    '''
    response = client.get(path, **kwargs)
    response_json = response.json()
    print(f'{response_json=}')
    assert response.status_code == expected_status_code
    return response_json


def add_entries(
    client,
    entries: Dict,
    access_token: auth.AccessToken,
):
    for resource, _entries in entries.items():
        for entry in _entries:
            response = client.post(
                f'/entries/{resource}/add',
                json={'entry': entry},
                headers=access_token.as_header(),
            )
            assert response.status_code == 201, f'Response(status_code={response.status_code}, json={response.json()})'
    return client



from .search_service import (
    EntryDto,
    SearchService,
    QueryRequest,
    QueryResponse,
    QuerySplitResponse,
    StatisticsDto,
)
from .resources import ResourceViews
from karp.search.application.queries.entries import (
    PreviewEntry, PreviewEntryInputDto, EntryPreviewDto)



#!/usr/bin/env python

# CAVEAT UTILITOR
#
# This file was automatically generated by TatSu.
#
#    https://pypi.python.org/pypi/tatsu/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.

from __future__ import annotations

import sys

from tatsu.buffering import Buffer
from tatsu.parsing import Parser
from tatsu.parsing import tatsumasu
from tatsu.parsing import leftrec, nomemo, isname # noqa
from tatsu.infos import ParserConfig
from tatsu.util import re, generic_main  # noqa


KEYWORDS = {}  # type: ignore


class KarpQueryV6Buffer(Buffer):
    def __init__(self, text, /, config: ParserConfig = None, **settings):
        base_config = ParserConfig.new(
            owner=self,
            whitespace=None,
            nameguard=None,
            comments_re=None,
            eol_comments_re=None,
            ignorecase=False,
            namechars='',
            parseinfo=False,
        )
        config = base_config.replace_config(config)
        config = config.merge(**settings)
        super().__init__(text, config=config)


class KarpQueryV6Parser(Parser):
    def __init__(self, config: ParserConfig = None, **settings):
        base_config = ParserConfig.new(
            owner=self,
            whitespace=None,
            nameguard=None,
            comments_re=None,
            eol_comments_re=None,
            ignorecase=False,
            namechars='',
            parseinfo=False,
            keywords=KEYWORDS,
        )
        config = base_config.replace_config(config)
        config = config.merge(**settings)
        super().__init__(config=config)

    @tatsumasu()
    def _start_(self):  # noqa
        self._expression_()
        self._check_eof()

    @tatsumasu()
    def _expression_(self):  # noqa
        with self._choice():
            with self._option():
                self._logical_expression_()
            with self._option():
                self._query_expression_()
            self._error(
                'expecting one of: '
                '<and> <or> <not> <logical_expression>'
                '<contains> <endswith> <equals> <exists>'
                '<freergxp> <freetext> <gt> <gte> <lt>'
                '<lte> <missing> <regexp> <startswith>'
                '<query_expression>'
            )

    @tatsumasu()
    def _query_expression_(self):  # noqa
        with self._choice():
            with self._option():
                self._contains_()
            with self._option():
                self._endswith_()
            with self._option():
                self._equals_()
            with self._option():
                self._exists_()
            with self._option():
                self._freergxp_()
            with self._option():
                self._freetext_()
            with self._option():
                self._gt_()
            with self._option():
                self._gte_()
            with self._option():
                self._lt_()
            with self._option():
                self._lte_()
            with self._option():
                self._missing_()
            with self._option():
                self._regexp_()
            with self._option():
                self._startswith_()
            self._error(
                'expecting one of: '
                "'contains' 'endswith' 'equals' 'exists'"
                "'freergxp' <freetext_any>"
                "<freetext_string> <freetext> 'gt' 'gte'"
                "'lt' 'lte' 'missing' 'regexp'"
                "'startswith'"
            )

    @tatsumasu()
    def _logical_expression_(self):  # noqa
        with self._choice():
            with self._option():
                self._and_()
            with self._option():
                self._or_()
            with self._option():
                self._not_()
            self._error(
                'expecting one of: '
                "'and' 'or' 'not'"
            )

    @tatsumasu('And')
    def _and_(self):  # noqa
        self._token('and')
        self.name_last_node('op')
        self._token('(')
        self._expression_()
        self.add_last_node_to_name('exps')

        def block2():
            self._token('||')
            self._expression_()
            self.add_last_node_to_name('exps')
        self._closure(block2)
        self._token(')')
        self._define(
            ['op'],
            ['exps']
        )

    @tatsumasu('Contains')
    def _contains_(self):  # noqa
        self._token('contains')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('field')
        self._token('|')
        self._string_()
        self.name_last_node('arg')
        self._define(
            ['arg', 'field', 'op'],
            []
        )

    @tatsumasu('Endswith')
    def _endswith_(self):  # noqa
        self._token('endswith')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('field')
        self._token('|')
        self._string_()
        self.name_last_node('arg')
        self._define(
            ['arg', 'field', 'op'],
            []
        )

    @tatsumasu('Equals')
    def _equals_(self):  # noqa
        self._token('equals')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('field')
        self._token('|')
        self._argument_()
        self.name_last_node('arg')
        self._define(
            ['arg', 'field', 'op'],
            []
        )

    @tatsumasu('Exists')
    def _exists_(self):  # noqa
        self._token('exists')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('field')
        self._define(
            ['field', 'op'],
            []
        )

    @tatsumasu('Freergxp')
    def _freergxp_(self):  # noqa
        self._token('freergxp')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('arg')
        self._define(
            ['arg', 'op'],
            []
        )

    @tatsumasu()
    def _freetext_(self):  # noqa
        with self._choice():
            with self._option():
                self._freetext_any_()
            with self._option():
                self._freetext_string_()
            self._error(
                'expecting one of: '
                "'freetext' <freetext_any>"
                '<freetext_string>'
            )

    @tatsumasu('FreetextAnyButString')
    def _freetext_any_(self):  # noqa
        self._token('freetext')
        self.name_last_node('op')
        self._token('|')
        self._any_but_string_()
        self.name_last_node('arg')
        self._define(
            ['arg', 'op'],
            []
        )

    @tatsumasu('FreetextString')
    def _freetext_string_(self):  # noqa
        self._token('freetext')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('arg')
        self._define(
            ['arg', 'op'],
            []
        )

    @tatsumasu('Gt')
    def _gt_(self):  # noqa
        self._token('gt')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('field')
        self._token('|')
        self._argument_()
        self.name_last_node('arg')
        self._define(
            ['arg', 'field', 'op'],
            []
        )

    @tatsumasu('Gte')
    def _gte_(self):  # noqa
        self._token('gte')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('field')
        self._token('|')
        self._argument_()
        self.name_last_node('arg')
        self._define(
            ['arg', 'field', 'op'],
            []
        )

    @tatsumasu('Lt')
    def _lt_(self):  # noqa
        self._token('lt')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('field')
        self._token('|')
        self._argument_()
        self.name_last_node('arg')
        self._define(
            ['arg', 'field', 'op'],
            []
        )

    @tatsumasu('Lte')
    def _lte_(self):  # noqa
        self._token('lte')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('field')
        self._token('|')
        self._argument_()
        self.name_last_node('arg')
        self._define(
            ['arg', 'field', 'op'],
            []
        )

    @tatsumasu('Missing')
    def _missing_(self):  # noqa
        self._token('missing')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('field')
        self._define(
            ['field', 'op'],
            []
        )

    @tatsumasu('Not')
    def _not_(self):  # noqa
        self._token('not')
        self.name_last_node('op')
        self._token('(')
        self._expression_()
        self.add_last_node_to_name('exps')

        def block2():
            self._token('||')
            self._expression_()
            self.add_last_node_to_name('exps')
        self._closure(block2)
        self._token(')')
        self._define(
            ['op'],
            ['exps']
        )

    @tatsumasu('Or')
    def _or_(self):  # noqa
        self._token('or')
        self.name_last_node('op')
        self._token('(')
        self._expression_()
        self.add_last_node_to_name('exps')

        def block2():
            self._token('||')
            self._expression_()
            self.add_last_node_to_name('exps')
        self._closure(block2)
        self._token(')')
        self._define(
            ['op'],
            ['exps']
        )

    @tatsumasu('Regexp')
    def _regexp_(self):  # noqa
        self._token('regexp')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('field')
        self._token('|')
        self._string_()
        self.name_last_node('arg')
        self._define(
            ['arg', 'field', 'op'],
            []
        )

    @tatsumasu('Startswith')
    def _startswith_(self):  # noqa
        self._token('startswith')
        self.name_last_node('op')
        self._token('|')
        self._string_()
        self.name_last_node('field')
        self._token('|')
        self._string_()
        self.name_last_node('arg')
        self._define(
            ['arg', 'field', 'op'],
            []
        )

    @tatsumasu()
    def _argument_(self):  # noqa
        with self._choice():
            with self._option():
                self._integer_()
            with self._option():
                self._string_()
            self._error(
                'expecting one of: '
                '\\d+ <integer> [^|()]+ <string>'
            )

    @tatsumasu()
    def _any_but_string_(self):  # noqa
        self._integer_()

    @tatsumasu()
    def _string_(self):  # noqa
        self._pattern('[^|()]+')

    @tatsumasu('int')
    def _integer_(self):  # noqa
        self._pattern('\\d+')


class KarpQueryV6Semantics(object):
    def start(self, ast):  # noqa
        return ast

    def expression(self, ast):  # noqa
        return ast

    def query_expression(self, ast):  # noqa
        return ast

    def logical_expression(self, ast):  # noqa
        return ast

    def and_(self, ast):  # noqa
        return ast

    def contains(self, ast):  # noqa
        return ast

    def endswith(self, ast):  # noqa
        return ast

    def equals(self, ast):  # noqa
        return ast

    def exists(self, ast):  # noqa
        return ast

    def freergxp(self, ast):  # noqa
        return ast

    def freetext(self, ast):  # noqa
        return ast

    def freetext_any(self, ast):  # noqa
        return ast

    def freetext_string(self, ast):  # noqa
        return ast

    def gt(self, ast):  # noqa
        return ast

    def gte(self, ast):  # noqa
        return ast

    def lt(self, ast):  # noqa
        return ast

    def lte(self, ast):  # noqa
        return ast

    def missing(self, ast):  # noqa
        return ast

    def not_(self, ast):  # noqa
        return ast

    def or_(self, ast):  # noqa
        return ast

    def regexp(self, ast):  # noqa
        return ast

    def startswith(self, ast):  # noqa
        return ast

    def argument(self, ast):  # noqa
        return ast

    def any_but_string(self, ast):  # noqa
        return ast

    def string(self, ast):  # noqa
        return ast

    def integer(self, ast):  # noqa
        return ast


def main(filename, start=None, **kwargs):
    if start is None:
        start = 'start'
    if not filename or filename == '-':
        text = sys.stdin.read()
    else:
        with open(filename) as f:
            text = f.read()
    parser = KarpQueryV6Parser()
    return parser.parse(
        text,
        rule_name=start,
        filename=filename,
        **kwargs
    )


if __name__ == '__main__':
    import json
    from tatsu.util import asjson

    ast = generic_main(main, KarpQueryV6Parser, name='KarpQueryV6')
    data = asjson(ast)
    print(json.dumps(data, indent=2))




import abc
import typing
from typing import Dict, List, Optional, Tuple

from karp.foundation import entity, events, repository, unit_of_work
from karp.lex.domain import entities, errors
from karp.lex.domain.value_objects import UniqueId


class EntryRepository(repository.Repository[entities.Entry]):
    EntityNotFound = errors.EntryNotFound

    @classmethod
    @abc.abstractmethod
    def from_dict(cls, settings: Dict):
        raise NotImplementedError()

    @classmethod
    @abc.abstractmethod
    def _create_repository_settings(
        cls, resource_id: str, resource_config: typing.Dict
    ):
        raise NotImplementedError()

    def __init__(self):
        super().__init__()
        self.settings = {}

    def by_id(
        self,
        id_: UniqueId,
        *,
        version: Optional[int] = None,
        after_date: Optional[float] = None,
        before_date: Optional[float] = None,
        oldest_first: bool = False,
        **kwargs,
    ) -> entities.Entry:
        entry = self._by_id(
            id_,
            version=version,
            after_date=after_date,
            before_date=before_date,
            oldest_first=oldest_first,
        )
        if entry:
            self.seen.add(entry)
            return entry
        raise errors.EntryNotFound(id_=id_)

    @abc.abstractmethod
    def _by_id(
        self,
        id: UniqueId,
        *,
        version: Optional[int] = None,
        after_date: Optional[float] = None,
        before_date: Optional[float] = None,
        oldest_first: bool = False,
        **kwargs,
    ) -> typing.Optional[entities.Entry]:
        raise NotImplementedError()

    # # @abc.abstractmethod
    # def move(self, entry: entities.Entry, *, old_entry_id: str):
    #     raise NotImplementedError()

    # # @abc.abstractmethod
    # def delete(self, entry: entities.Entry):
    #     raise NotImplementedError()

    # @abc.abstractmethod
    # def entry_ids(self) -> List[str]:
    # raise NotImplementedError()

    def entity_ids(self) -> List[str]:
        raise NotImplementedError()

    # def by_entry_id(
    #     self, entry_id: str, *, version: Optional[int] = None
    # ) -> entities.Entry:
    #     entry = self.get_by_entry_id_optional(
    #         entry_id,
    #         version=version,
    #     )
    #     if not entry:
    #         raise errors.EntryNotFound(
    #             f'Entry with entry_id="{entry_id}"',
    #             entity_id=None,
    #         )
    #     return entry

    # get_by_entry_id = by_entry_id

    # def get_by_entry_id_optional(
    #     self,
    #     entry_id: str,
    #     *,
    #     version: Optional[int] = None,
    # ) -> Optional[entities.Entry]:
    #     entry = self._by_entry_id(entry_id)
    #     if not entry:
    #         return None
    #     if version:
    #         entry = self._by_id(entry.entity_id, version=version)
    #     if entry:
    #         self.seen.add(entry)
    #         return entry
    #     return None

    # @abc.abstractmethod
    # def _by_entry_id(
    #     self,
    #     entry_id: str,
    # ) -> Optional[entities.Entry]:
    #     raise NotImplementedError()

    # @abc.abstractmethod
    def teardown(self):
        """Use for testing purpose."""
        return

    @abc.abstractmethod
    def by_referenceable(
        self, filters: Optional[Dict] = None, **kwargs
    ) -> list[entities.Entry]:
        raise NotImplementedError()

    # @abc.abstractmethod
    def get_history(
        self,
        user_id: Optional[str] = None,
        entry_id: Optional[str] = None,
        from_date: Optional[float] = None,
        to_date: Optional[float] = None,
        from_version: Optional[int] = None,
        to_version: Optional[int] = None,
        offset: int = 0,
        limit: int = 100,
    ) -> Tuple[List[entities.Entry], int]:
        return [], 0

    @abc.abstractmethod
    def all_entries(self) -> typing.Iterable[entities.Entry]:
        """Return all entries."""
        return []

    def get_total_entries(self) -> int:
        return len(list(self.all_entries()))


class EntryUnitOfWork(
    unit_of_work.UnitOfWork[EntryRepository],
    entity.TimestampedEntity,
):
    repository_type: str

    def __init__(
        self,
        name: str,
        config: Dict,
        connection_str: Optional[str],
        message: str,
        event_bus: events.EventBus,
        *args,
        **kwargs,
    ):
        unit_of_work.UnitOfWork.__init__(self, event_bus)
        entity.TimestampedEntity.__init__(self, *args, **kwargs)
        self._name = name
        self._connection_str = connection_str
        self._config = config
        self._message = message

    @property
    def entries(self) -> EntryRepository:
        return self.repo

    @property
    def name(self) -> str:
        return self._name

    @property
    def connection_str(self) -> Optional[str]:
        return self._connection_str

    @property
    def config(self) -> Dict:
        return self._config

    @property
    def message(self) -> str:
        return self._message

    def discard(self, *, user, timestamp: Optional[float] = None):
        self._discarded = True
        self._last_modified = self._ensure_timestamp(timestamp)
        self._last_modified_by = user
        # return entity.TimestampedEntity.discard(self, user=user, timestamp=timestamp)



from dataclasses import dataclass

from karp.lex.domain.value_objects import UniqueId


@dataclass
class ResourceReference:
    resource_id: UniqueId


@dataclass
class ResourceSoftReference:
    resource_id: str


@dataclass
class EntryReference:
    resource_id: UniqueId
    entry_id: UniqueId


@dataclass
class EntrySoftReference:
    resource_id: str
    entry_id: str



import logging
from pathlib import Path
import threading
from typing import Dict, Type
import sys

try:
    from importlib.metadata import entry_points
except ImportError:
    # used if python == 3.9
    from importlib_metadata import entry_points  # type: ignore

import elasticsearch
import injector
from injector import Provider, T
from sqlalchemy.engine import Connection, Engine
from sqlalchemy.orm import Session, sessionmaker

from karp.foundation.commands import CommandBus, InjectorCommandBus
from karp.foundation.events import EventBus, InjectorEventBus
from karp.auth_infrastructure import AuthInfrastructure, TestAuthInfrastructure, JwtAuthInfrastructure


# from foundation.events import EventBus, InjectorEventBus, RunAsyncHandler
# from foundation.locks import Lock, LockFactory
#
# from customer_relationship import CustomerRelationshipConfig
# from main.async_handler_task import async_handler_generic_task
# from main.redis import RedisLock
# from payments import PaymentsConfig


logger = logging.getLogger(__name__)


class RequestScope(injector.Scope):
    REGISTRY_KEY = "RequestScopeRegistry"

    def configure(self) -> None:
        self._locals = threading.local()

    def enter(self) -> None:
        logger.warning('entering request scope')
        assert not hasattr(self._locals, self.REGISTRY_KEY)
        setattr(self._locals, self.REGISTRY_KEY, {})

    def exit(self) -> None:
        logger.warning('exiting request scope')
        for key, provider in getattr(self._locals, self.REGISTRY_KEY).items():
            provider.get(self.injector).close()
            delattr(self._locals, repr(key))

        delattr(self._locals, self.REGISTRY_KEY)

    def __enter__(self) -> None:
        self.enter()

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:  # type: ignore
        self.exit()

    def get(self, key: Type[T], provider: Provider[T]) -> Provider[T]:
        try:
            return getattr(self._locals, repr(key))  # type: ignore
        except AttributeError:
            provider = injector.InstanceProvider(provider.get(self.injector))
            setattr(self._locals, repr(key), provider)
            try:
                registry = getattr(self._locals, self.REGISTRY_KEY)
            except AttributeError:
                raise Exception(
                    f"{key} is request scoped, but no RequestScope entered!")
            registry[key] = provider
            return provider


request = injector.ScopeDecorator(RequestScope)


class CommandBusMod(injector.Module):
    @injector.provider
    def command_bus(self, inj: injector.Injector) -> CommandBus:
        return InjectorCommandBus(inj)


class EventBusMod(injector.Module):
    @injector.provider
    def event_bus(self, inj: injector.Injector) -> EventBus:
        return InjectorEventBus(inj)


class Db(injector.Module):
    def __init__(self, engine: Engine) -> None:
        self._engine = engine

    # @request
    @injector.provider
    def connection(self) -> Connection:
        return self._engine.connect()

    # @request
    @injector.provider
    def session(self, connection: Connection) -> Session:
        return Session(bind=connection)

    # @request
    @injector.provider
    def session_factory(self) -> sessionmaker:
        return sessionmaker(bind=self._engine)


class ElasticSearchMod(injector.Module):
    def __init__(self, es_url: str) -> None:
        self._url = es_url

    @injector.provider
    @injector.singleton
    def es(self) -> elasticsearch.Elasticsearch:
        logger.info('Creating ES client url=%s', self._url)
        return elasticsearch.Elasticsearch(self._url)


TEST_AUTH_SERVICE = 'DUMMY_AUTH'


def install_auth_service(
    container: injector.Injector,
    settings: Dict[str, str]
):
    auth_service_name = settings.get('auth.name', '')
    container.binder.install(AuthInfrastructure())

    if auth_service_name.upper() == TEST_AUTH_SERVICE:
        container.binder.install(TestAuthInfrastructure())
    else:
        container.binder.install(JwtAuthInfrastructure(
            Path(settings['auth.jwt.pubkey.path'])))


def request_configuration(conn: Connection, session: Session):
    def configure_request_container(binder):
        binder.bind(Connection, to=injector.InstanceProvider(conn))
        binder.bind(Session, to=injector.InstanceProvider(session))

    return configure_request_container


def load_modules(group_name: str, app=None):
    logger.debug('Loading %s', group_name)
    if sys.version_info.minor < 10:
        karp_modules = entry_points().get(group_name)
    else:
        karp_modules = entry_points(group=group_name)  # type: ignore
    if karp_modules:
        for ep in karp_modules:
            logger.info("Loading '%s' from '%s'", ep.name, group_name, extra={
                        'group_name': group_name, 'entry_point_name': ep.name})
            mod = ep.load()
            if app:
                init_app = getattr(mod, "init_app", None)
                if init_app:
                    init_app(app)



import logging
import typing

from fastapi import APIRouter, Body, Depends, HTTPException, Security, status
import logging

from karp import auth, lex
from karp.auth.application.queries import GetResourcePermissions, ResourcePermissionDto
from karp.foundation.commands import CommandBus

from karp.webapp.schemas import ResourceCreate, ResourcePublic, ResourceProtected
from karp.webapp import dependencies as deps, schemas

from karp.lex import (
    CreatingEntryRepo,
    CreatingResource,
    ResourceDto,
)
from karp.lex.domain import commands
from karp.lex.application.queries import ReadOnlyResourceRepository


router = APIRouter()
logger = logging.getLogger(__name__)


@router.get("/permissions", response_model=list[ResourcePermissionDto])
def list_resource_permissions(
    query: GetResourcePermissions = Depends(deps.get_resource_permissions),
):
    return query.query()


@router.get(
    "/",
    response_model=list[ResourceProtected],
)
def get_all_resources(
    get_resources: lex.GetResources = Depends(deps.inject_from_req(lex.GetResources)),
) -> typing.Iterable[lex.ResourceDto]:
    return get_resources.query()


@router.post("/", response_model=ResourceDto, status_code=status.HTTP_201_CREATED)
def create_new_resource(
    new_resource: ResourceCreate = Body(...),
    user: auth.User = Security(deps.get_user, scopes=["admin"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    creating_resource_uc: CreatingResource = Depends(deps.get_lex_uc(CreatingResource)),
    creating_entry_repo_uc: CreatingEntryRepo = Depends(
        deps.get_lex_uc(CreatingEntryRepo)
    ),
) -> ResourceDto:
    logger.info(
        "creating new resource",
        extra={"user": user.identifier, "resource": new_resource},
    )
    if not auth_service.authorize(
        auth.PermissionLevel.admin, user, [new_resource.resource_id]
    ):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="lexica:admin"'},
        )
    try:
        if new_resource.entry_repo_id is None:
            entry_repo = creating_entry_repo_uc.execute(
                commands.CreateEntryRepository(
                    repository_type="default",
                    name=new_resource.resource_id,
                    config=new_resource.config,
                    user=user.identifier,
                    message=new_resource.message or "Entry repository created",
                )
            )
            new_resource.entry_repo_id = entry_repo.entity_id
        create_resource = commands.CreateResource(
            user=user.identifier,
            **new_resource.dict(),
        )
        resource = creating_resource_uc.execute(create_resource)
        logger.info("resource created", extra={"resource": resource})
        return resource
    except Exception as err:
        logger.exception(
            "error occured", extra={"user": user.identifier, "resource": new_resource}
        )
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"{err=}",
        )


@router.post(
    "/{resource_id}/publish",
    # response_model=ResourceDto,
    status_code=status.HTTP_200_OK,
)
def publishing_resource(
    resource_id: str,
    resource_publish: schemas.ResourcePublish = Body(...),
    user: auth.User = Security(deps.get_user, scopes=["admin"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    publishing_resource_uc: lex.PublishingResource = Depends(
        deps.get_lex_uc(lex.PublishingResource)
    ),
):
    if not auth_service.authorize(auth.PermissionLevel.admin, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="lexica:admin"'},
        )
    logger.info(
        "publishing resource",
        extra={"resource_id": resource_id, "user": user.identifier},
    )
    try:
        resource_publish.resource_id = resource_id
        publish_resource = commands.PublishResource(
            user=user.identifier,
            **resource_publish.dict(),
        )
        publishing_resource_uc.execute(publish_resource)
        logger.info("resource published", extra={"resource_id": resource_id})
        return
    except Exception as err:
        logger.exception(
            "error occured when publishing",
            extra={"resource_id": resource_id, "user": user.identifier},
        )
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"{err=}",
        ) from err


@router.get(
    "/{resource_id}",
    response_model=ResourcePublic,
)
def get_resource_by_resource_id(
    resource_id: str,
    resource_repo: ReadOnlyResourceRepository = Depends(deps.get_resources_read_repo),
) -> ResourcePublic:
    resource = resource_repo.get_by_resource_id(resource_id)
    if not resource:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"No resource with resource_id '{resource_id}' was found.",
        )
    return resource


def init_app(app):
    app.include_router(router)



import abc
import typing

import pydantic

from karp.foundation import repository, unit_of_work


class IndexEntry(pydantic.BaseModel):
    id: str
    entry: typing.Dict

    def __bool__(self) -> bool:
        return bool(self.entry)


class Index(repository.Repository[IndexEntry]):

    @abc.abstractmethod
    def create_index(self, resource_id: str, config: typing.Dict):
        pass

    @abc.abstractmethod
    def publish_index(self, resource_id: str):
        pass

    @abc.abstractmethod
    def add_entries(self, resource_id: str, entries: typing.Iterable[IndexEntry]):
        pass

    @abc.abstractmethod
    def delete_entry(
        self,
        resource_id: str,
        *,
        entry_id: typing.Optional[str] = None,
    ):
        pass

    def create_empty_object(self) -> IndexEntry:
        return IndexEntry(id='', entry={})

    def assign_field(self, _index_entry: IndexEntry, field_name: str, part):
        if isinstance(part, IndexEntry):
            part = part.entry
        _index_entry.entry[field_name] = part

    def create_empty_list(self) -> typing.List:
        return []

    def add_to_list_field(self, elems: typing.List, elem):
        if isinstance(elem, IndexEntry):
            elem = elem.entry
        elems.append(elem)

    def _save(self, _notused):
        pass

    def _by_id(self, id) -> None:
        return None

    def num_entities(self) -> int:
        raise NotImplementedError("num_entities is not used for indicies")


class IndexUnitOfWork(
    unit_of_work.UnitOfWork[Index]
):
    pass



import abc
import logging
import typing

from karp.foundation.events import EventBus


RepositoryType = typing.TypeVar("RepositoryType")


logger = logging.getLogger(__name__)


class UnitOfWork(typing.Generic[RepositoryType], abc.ABC):
    def __init__(self, event_bus: EventBus) -> None:
        self.event_bus = event_bus

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.rollback()
        self.close()

    def commit(self):
        logger.debug('called commit')
        self._commit()
        for event in self.collect_new_events():
            logger.debug('collecting events')
            self.event_bus.post(event)

    def collect_new_events(self) -> typing.Iterable:
        for entity in self.repo.seen:
            # yield from entity.domain_events
            yield from entity.collect_new_events()
            entity.clear_events()

    @abc.abstractmethod
    def _commit(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def rollback(self):
        pass

    @property
    @abc.abstractmethod
    def repo(self) -> RepositoryType:
        pass

    @abc.abstractmethod
    def _close(self):
        pass

    def close(self):
        self._close()



import logging
from typing import Any

from karp.foundation.commands import CommandHandler
from karp.lex.application.repositories.entries import EntryUnitOfWork
from karp.lex.domain import commands
from karp.lex.application import repositories


logger = logging.getLogger(__name__)


class CreatingEntryRepo(CommandHandler[commands.CreateEntryRepository]):
    def __init__(
        self,
        entry_repo_uow: repositories.EntryUowRepositoryUnitOfWork,
        **kwargs,
    ):
        self._entry_repo_uow = entry_repo_uow

    def execute(self, command: commands.CreateEntryRepository) -> EntryUnitOfWork:
        entry_repo = self._entry_repo_uow.factory.create(
            repository_type=command.repository_type,
            entity_id=command.entity_id,
            name=command.name,
            config=command.config,
            connection_str=command.connection_str,
            user=command.user,
            timestamp=command.timestamp,
            message=command.message,
        )

        logger.debug("Created entry repo", extra={"entry_repo": entry_repo})
        with self._entry_repo_uow as uow:
            logger.debug("Saving...")
            uow.repo.save(entry_repo)
            uow.commit()
        return entry_repo


class DeletingEntryRepository(CommandHandler[commands.DeleteEntryRepository]):
    def __init__(
        self,
        entry_repo_uow: repositories.EntryUowRepositoryUnitOfWork,
        **kwargs,
    ):
        self._entry_repo_uow = entry_repo_uow

    def execute(self, command: commands.DeleteEntryRepository) -> None:
        with self._entry_repo_uow as uow:
            entry_repo = uow.repo.get_by_id(command.entity_id)
            entry_repo.discard(user=command.user, timestamp=command.timestamp)
            uow.repo.save(entry_repo)
            uow.commit()



from pathlib import Path
from typing import Optional

import json_streams
import typer

from karp import search
from karp.cliapp.typer_injector import inject_from_ctx


subapp = typer.Typer()


@subapp.command()
def resource(
    ctx: typer.Context,
    resource_id: str,
    output: Optional[Path] = typer.Option(
        None, help="Path to write to. Defaults to stdout."
    ),
):
    search_service = inject_from_ctx(search.SearchService, ctx)
    query_request = search.QueryRequest(resource_ids=[resource_id])
    json_streams.dump_to_file(
        search_service.query(query_request),
        output,
        use_stdout_as_default=True,
    )


def init_app(app: typer.Typer) -> None:
    app.add_typer(subapp, name="query")



def noswedish(s: str) -> str:
    result = ""
    for c in s:
        if c == "å":
            result += "aa"
        elif c == "ä":
            result += "ae"
        elif c == "ö":
            result += "oe"
        elif c == "Å":
            result += "Aa"
        elif c == "Ä":
            result += "Ae"
        elif c == "Ö":
            result += "Oe"
        else:
            result += c

    return result



"""LexicalResource"""
import abc
import enum
import typing
from typing import Any, Callable, Dict, List, Optional, Type, Union
from uuid import UUID

from karp.lex.domain import constraints
from karp.foundation.entity import Entity, TimestampedVersionedEntity
from karp.foundation.value_objects import PermissionLevel
from .entry import Entry, create_entry
from karp.lex.domain import errors, events
from karp.foundation.value_objects import unique_id
from karp.utility import json_schema, time
from karp.utility.container import create_field_getter
from karp.utility.time import utc_now

# pylint: disable=unsubscriptable-object


class ResourceOp(enum.Enum):
    ADDED = "ADDED"
    UPDATED = "UPDATED"
    DELETED = "DELETED"


class Resource(TimestampedVersionedEntity):
    DiscardedEntityError = errors.DiscardedEntityError
    resource_type: str = "resource"

    # @classmethod
    # def create_resource(cls, resource_type: str, resource_config: Dict):
    #     if resource_type == cls.resource_type:
    #         return cls.from_dict(resource_config)

    # @classmethod
    # def from_dict(cls, config: Dict, **kwargs):

    #     resource_id = config.pop("resource_id")
    #     resource_name = config.pop("resource_name")

    #     #     entry_repository = EntryRepository.create(
    #     #         config["entry_repository_type"],
    #     #         entry_repository_settings
    #     #     )

    #     resource = cls(
    #         resource_id=resource_id,
    #         name=resource_name,
    #         config=config,
    #         message="Resource added.",
    #         op=ResourceOp.ADDED,
    #         entity_id=unique_id.make_unique_id(),
    #         version=1,
    #         **kwargs,
    #     )
    #     resource.queue_event(
    #         events.ResourceCreated(
    #             entity_id=resource.id,
    #             id=resource.id,
    #             resource_id=resource.resource_id,
    #             name=resource.name,
    #             config=resource.config,
    #             timestamp=resource.last_modified,
    #             user=resource.last_modified_by,
    #             message=resource.message,
    #         )
    #     )
    #     return resource

    # class NewReleaseAdded:
    #     def mutate(self, obj):
    #         obj._validate_event_applicability(self)
    #         release = Release(
    #             entity_id=self.release_id,
    #             name=self.release_name,
    #             publication_date=self.timestamp,
    #             description=self.release_description,
    #             aggregate_root=obj,
    #         )
    #         obj._releases.append(release)
    #         obj._last_modified = self.timestamp
    #         obj._last_modified_by = self.user
    #         obj._message = f"Release '{self.release_name}' created."
    #         obj._increment_version()

    def __init__(
        self,
        *,
        entity_id: unique_id.UniqueId,
        resource_id: str,
        name: str,
        config: Dict[str, Any],
        message: str,
        entry_repo_id: unique_id.UniqueId,
        version: int = 1,
        op: ResourceOp = ResourceOp.ADDED,
        is_published: bool = False,
        # entry_repository_type: typing.Optional[str] = None,
        # entry_repository_settings: typing.Optional[typing.Dict] = None,
        # entry_repository: EntryRepository = None,
        **kwargs,
    ):
        super().__init__(entity_id=entity_id, version=version, **kwargs)
        self._resource_id = resource_id
        self._name = name
        self.is_published = is_published
        self.config = config
        self._message = message
        self._op = op
        self._releases = []
        # self._entry_repository = None
        # self.entry_repository_type = entry_repository_type
        # self.entry_repository_settings = entry_repository_settings
        self._entry_json_schema = None
        self._entry_repo_id = entry_repo_id
        # if not self.events or not isinstance(self.events[-1], events.ResourceCreated):
        #     self.queue_event(
        #         events.ResourceLoaded(
        #             id=self._id,
        #             resource_id=self._resource_id,
        #             name=self._name,
        #             config=self.config,
        #             timestamp=self._last_modified,
        #             user=self._last_modified_by,
        #             message=self._message,
        #             version=self.version,
        #         )
        #     )

    @property
    def resource_id(self) -> str:
        return self._resource_id

    @property
    def entry_repository_id(self) -> unique_id.UniqueId:
        return self._entry_repo_id

    @property
    def entry_repository_settings(self) -> typing.Dict:
        return self.config["entry_repository_settings"]

    @entry_repository_settings.setter
    def entry_repository_settings(self, entry_repository_settings: typing.Dict):
        self.config["entry_repository_settings"] = entry_repository_settings

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._check_not_discarded()
        self._name = name

    @property
    def message(self):
        return self._message

    @property
    def releases(self):
        """Releases for this resource."""
        return self._releases

    @property
    def op(self):
        return self._op

    # @property
    # def entry_repository(self):
    #     from karp.domain.repository import EntryRepository

    #     if self._entry_repository is None:
    #         self._entry_repository = EntryRepository.create(
    #             self.entry_repository_type,
    #             {"table_name": self._resource_id, "config": self.config},
    #         )
    #     return self._entry_repository

    # def stamp(
    #     self,
    #     *,
    #     user: str,
    #     timestamp: float = None,
    #     message: str = None,
    #     increment_version: bool = True,
    # ):
    #     self._update_metadata(timestamp, user, message, "Updated")
    #     if increment_version:
    #         self._version += 1
    #     self.queue_event(
    #         events.ResourceUpdated(
    #             entity_id=self.id,
    #             resource_id=self.resource_id,
    #             name=self.name,
    #             config=self.config,
    #             version=self.version,
    #             timestamp=self.last_modified,
    #             user=self.last_modified_by,
    #             message=self.message,
    #             entry_repo_id=self.entry_repository_id,
    #         )
    #     )

    def publish(
        self,
        *,
        user: str,
        message: str,
        version: int,
        timestamp: float = None,
    ):
        self._update_metadata(timestamp, user, message or "Published", version)
        self.is_published = True
        self.queue_event(
            events.ResourcePublished(
                entity_id=self.id,
                resource_id=self.resource_id,
                entry_repo_id=self.entry_repository_id,
                timestamp=self.last_modified,
                user=self.last_modified_by,
                version=self.version,
                name=self.name,
                config=self.config,
                message=self.message,
            )
        )

    def set_entry_repo_id(
        self,
        *,
        entry_repo_id: unique_id.UniqueId,
        user: str,
        version: int,
        timestamp: Optional[float] = None,
    ):
        self._update_metadata(timestamp, user, "entry repo id updated", version=version)
        self._entry_repo_id = entry_repo_id
        self.queue_event(
            events.ResourceUpdated(
                entity_id=self.entity_id,
                resource_id=self.resource_id,
                entry_repo_id=self.entry_repository_id,
                timestamp=self.last_modified,
                user=self.last_modified_by,
                version=self.version,
                name=self.name,
                config=self.config,
                message=self.message,
            )
        )

    def set_resource_id(
        self,
        *,
        resource_id: str,
        user: str,
        version: int,
        timestamp: Optional[float] = None,
        message: Optional[str] = None,
    ) -> None:
        self._update_metadata(
            timestamp, user, message or "setting resource_id", version
        )
        self._resource_id = resource_id
        self.queue_event(
            events.ResourceUpdated(
                entity_id=self.entity_id,
                resource_id=self.resource_id,
                entry_repo_id=self.entry_repository_id,
                timestamp=self.last_modified,
                user=self.last_modified_by,
                version=self.version,
                name=self.name,
                config=self.config,
                message=self.message,
            )
        )

    def update(
        self,
        *,
        name: str,
        config: dict[str, Any],
        user: str,
        version: int,
        timestamp: Optional[float] = None,
        message: Optional[str] = None,
    ) -> bool:
        if self.name == name and self.config == config:
            return False
        self._update_metadata(timestamp, user, message or "updating", version)
        self._name = name
        self.config = config
        self.queue_event(
            events.ResourceUpdated(
                entity_id=self.entity_id,
                resource_id=self.resource_id,
                entry_repo_id=self.entry_repository_id,
                timestamp=self.last_modified,
                user=self.last_modified_by,
                version=self.version,
                name=self.name,
                config=self.config,
                message=self.message,
            )
        )
        return True

    def set_config(
        self,
        *,
        config: dict[str, Any],
        user: str,
        version: int,
        timestamp: Optional[float] = None,
        message: Optional[str] = None,
    ) -> bool:
        if self.config == config:
            return False
        self._update_metadata(timestamp, user, message or "setting config", version)
        self.config = config
        self.queue_event(
            events.ResourceUpdated(
                entity_id=self.entity_id,
                resource_id=self.resource_id,
                entry_repo_id=self.entry_repository_id,
                timestamp=self.last_modified,
                user=self.last_modified_by,
                version=self.version,
                name=self.name,
                config=self.config,
                message=self.message,
            )
        )
        return True

    def _update_metadata(
        self, timestamp: Optional[float], user: str, message: str, version: int
    ):
        self._check_not_discarded()
        self._validate_version(version)
        self._last_modified = self._ensure_timestamp(timestamp)
        self._last_modified_by = user
        self._message = message
        self._op = ResourceOp.UPDATED
        self._increment_version()

    def add_new_release(self, *, name: str, user: str, description: str):
        self._check_not_discarded()
        raise NotImplementedError()
        # event = Resource.NewReleaseAdded(
        #     entity_id=self.id,
        #     entity_version=self.version,
        #     entity_last_modified=self.last_modified,
        #     release_id=unique_id.make_unique_id(),
        #     release_name=constraints.length_gt_zero("name", name),
        #     user=user,
        #     release_description=description,
        # )
        # event.mutate(self)

        # return self.release_with_name(name)

    def release_with_name(self, name: str):
        self._check_not_discarded()
        raise NotImplementedError()

    def discard(self, *, user: str, message: str, timestamp: float = None):
        self._check_not_discarded()
        self._op = ResourceOp.DELETED
        self._message = message or "Entry deleted."
        self._discarded = True
        self._last_modified_by = user
        self._last_modified = timestamp or utc_now()
        self._version += 1
        self.queue_event(
            events.ResourceDiscarded(
                entity_id=self.id,
                version=self.version,
                # entry_repo_id=self.entry_repository_id,
                timestamp=self.last_modified,
                user=self.last_modified_by,
                message=self.message,
                resource_id=self.resource_id,
                name=self.name,
                config=self.config,
            )
        )

    @property
    def entry_json_schema(self) -> Dict:
        if self._entry_json_schema is None:
            self._entry_json_schema = json_schema.create_entry_json_schema(
                self.config["fields"]
            )
        return self._entry_json_schema

    # @property
    # def id_getter(self) -> Callable[[Dict], str]:
    #     return create_field_getter(self.config["id"], str)

    def dict(self) -> Dict:
        return {
            "entity_id": self.entity_id,
            "resource_id": self.resource_id,
            "name": self.name,
            "version": self.version,
            "last_modified": self.last_modified,
            "last_modified_by": self.last_modified_by,
            "op": self.op,
            "message": self.message,
            "entry_repository_id": self.entry_repository_id,
            "is_published": self.is_published,
            "discarded": self.discarded,
            "resource_type": self.resource_type,
            "config": self.config,
        }

    def create_entry_from_dict(
        self,
        entry_raw: Dict,
        *,
        user: str,
        entity_id: unique_id.UniqueId,
        message: Optional[str] = None,
        timestamp: Optional[float] = None,
    ) -> Entry:
        self._check_not_discarded()
        # id_getter = self.id_getter()
        return create_entry(
            # id_getter(entry_raw),
            entry_raw,
            repo_id=self.entry_repository_id,
            # resource_id=self.resource_id,
            last_modified_by=user,
            message=message,
            entity_id=entity_id,
            last_modified=timestamp,
        )

    def is_protected(self, level: PermissionLevel):
        """
        Level can be READ, WRITE or ADMIN
        """
        protection = self.config.get("protected", {})
        return level == "WRITE" or level == "ADMIN" or protection.get("read")


# ===== Entities =====
class Release(Entity):
    def __init__(self, name: str, publication_date: float, description: str, **kwargs):
        super().__init__(**kwargs)
        self._name = name
        self._publication_date = publication_date
        self._description = description

    @property
    def name(self) -> str:
        """The name of this release."""
        return self._name

    @property
    def publication_date(self) -> float:
        """The publication of this release."""
        return self._publication_date

    @property
    def description(self) -> str:
        """The description of this release."""
        return self._description


# ===== Factories =====


def create_resource(
    config: Dict,
    entry_repo_id: unique_id.UniqueId,
    created_by: str = None,
    user: str = None,
    created_at: float = None,
    entity_id: unique_id.UniqueId = None,
    resource_id: typing.Optional[str] = None,
    message: typing.Optional[str] = None,
    name: typing.Optional[str] = None,
) -> Resource:

    resource_id_in_config = config.pop("resource_id", None)
    resource_id = resource_id or resource_id_in_config
    if not resource_id:
        raise ValueError("resource_id is missing")
    resource_id = constraints.valid_resource_id(resource_id)
    name_in_config = config.pop("resource_name", None)
    resource_name = name or name_in_config or resource_id
    entity_id = entity_id or unique_id.make_unique_id()
    #
    #     entry_repository = EntryRepository.create(
    #         config["entry_repository_type"],
    #         entry_repository_settings
    #     )

    resource = Resource(
        entity_id=entity_id,
        resource_id=resource_id,
        name=resource_name,
        config=config,
        entry_repo_id=entry_repo_id,
        message=message or "Resource added.",
        op=ResourceOp.ADDED,
        version=1,
        last_modified=created_at or time.utc_now(),
        last_modified_by=user or created_by or "unknown",
    )
    resource.queue_event(
        events.ResourceCreated(
            entity_id=resource.id,
            resource_id=resource.resource_id,
            entry_repo_id=resource.entry_repository_id,
            name=resource.name,
            config=resource.config,
            timestamp=resource.last_modified,
            user=resource.last_modified_by,
            message=resource.message,
        )
    )
    return resource


# ===== Repository =====



import abc
import logging
import typing
from typing import Callable, Dict, List, Optional, Tuple, TypeVar

import pydantic

from karp.search.domain.query import Query

logger = logging.getLogger(__name__)


class EntryDto(pydantic.BaseModel):
    id: str
    version: int
    last_modified: float
    last_modified_by: str
    resource: str
    entry: typing.Dict


class QueryResponseBase(pydantic.BaseModel):
    hits: list[EntryDto]
    total: int
    distribution: typing.Optional[dict[str, int]]


class QueryResponse(QueryResponseBase):
    pass


class QuerySplitResponse(QueryResponseBase):
    distribution: dict[str, int]


class StatisticsDto(pydantic.BaseModel):
    value: str
    count: int


class QueryRequest(pydantic.BaseModel):  # pylint: disable=no-member
    resource_ids: typing.List[str]
    q: typing.Optional[str] = None
    from_: int = 0
    size: int = 25
    lexicon_stats: bool = True
    sort: List[str] = pydantic.Field(default_factory=list)

    @pydantic.validator("resource_ids", pre=True)
    @classmethod
    def split_str(cls, v):
        if isinstance(v, str):
            return v.split(",")
        return v


class SearchService(abc.ABC):

    # def build_query(self, args, resource_str: str) -> Query:
    #     query = Query()
    #     query.parse_arguments(args, resource_str)
    #     return query

    # def build_query_parsed(self, args, resource_str: str) -> Query:
    #     query = Query()
    #     query.parse_arguments(args, resource_str)
    #     return query

    # def search_with_query(self, query: Query):
    #     raise NotImplementedError()

    @abc.abstractmethod
    def search_ids(self, resource_id: str, entry_ids: str):
        raise NotImplementedError()

    @abc.abstractmethod
    def query(self, request: QueryRequest):
        raise NotImplementedError()

    @abc.abstractmethod
    def query_split(self, request: QueryRequest):
        raise NotImplementedError()

    @abc.abstractmethod
    def statistics(self, resource_id: str, field: str) -> typing.Iterable[StatisticsDto]:
        raise NotImplementedError()



from sqlalchemy.engine.url import URL, make_url


def sqlalchemy_url(db_url: str, driver: str = None, host: str = None, port: int = None, user: str = None, password: str = None, database: str = None) -> URL:
    if db_url:
        return make_url(db_url)
    return URL.create(drivername=driver, username=user, password=password, host=host, port=port, database=database)



from .entries import EntryRepository, EntryUnitOfWork
from .entry_repositories import (
    EntryUowRepository,
    EntryUowRepositoryUnitOfWork,
    EntryRepositoryUnitOfWorkFactory,
    InjectorEntryUnitOfWorkRepoFactory,
    EntryUnitOfWorkCreator,
    EntryUnitOfWork,
)
from .resources import ResourceRepository, ResourceUnitOfWork
# from .unit_of_work import EntriesUnitOfWork, ResourceUnitOfWork, EntryUnitOfWork, EntryUowFactory
# from .entry_repository_repository import EntryRepositoryRepositoryUnitOfWork



import logging
from typing import Optional
import uuid

from fastapi import (
    APIRouter,
    Body,
    Depends,
    HTTPException,
    Response,
    Security,
    status,
    Path,
    Query,
)
import pydantic
from starlette import responses
from karp.foundation.value_objects import unique_id
from karp.foundation.value_objects.unique_id import UniqueIdStr

from karp.lex.domain.value_objects import UniqueId
from karp import errors as karp_errors, auth, lex, search
from karp.lex.application.queries import EntryDto, GetEntryHistory
from karp.lex.domain import commands, errors
from karp.auth import User
from karp.foundation.value_objects import PermissionLevel
from karp.auth import AuthService
from karp.webapp import schemas, dependencies as deps


router = APIRouter()

logger = logging.getLogger(__name__)


@router.get(
    "/{resource_id}/{entry_id}/{version}", response_model=EntryDto, tags=["History"]
)
@router.get("/{resource_id}/{entry_id}", response_model=EntryDto, tags=["History"])
def get_history_for_entry(
    resource_id: str,
    entry_id: UniqueIdStr,
    version: Optional[int] = Query(None),
    user: auth.User = Security(deps.get_user, scopes=["admin"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    get_entry_history: GetEntryHistory = Depends(deps.get_entry_history),
):
    if not auth_service.authorize(auth.PermissionLevel.admin, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="lexica:admin"'},
        )
    logger.info(
        "getting history for entry",
        extra={
            "resource_id": resource_id,
            "entry_id": entry_id,
            "user": user.identifier,
        },
    )
    return get_entry_history.query(resource_id, entry_id, version=version)


@router.post(
    "/{resource_id}/add",
    status_code=status.HTTP_201_CREATED,
    tags=["Editing"],
    response_model=schemas.EntryAddResponse,
)
@router.put(
    "/{resource_id}",
    status_code=status.HTTP_201_CREATED,
    tags=["Editing"],
    response_model=schemas.EntryAddResponse,
)
def add_entry(
    resource_id: str,
    data: schemas.EntryAdd,
    user: User = Security(deps.get_user, scopes=["write"]),
    auth_service: AuthService = Depends(deps.get_auth_service),
    adding_entry_uc: lex.AddingEntry = Depends(deps.get_lex_uc(lex.AddingEntry)),
):

    if not auth_service.authorize(PermissionLevel.write, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="write"'},
        )
    logger.info("adding entry", extra={"resource_id": resource_id, "data": data})
    try:
        new_entry = adding_entry_uc.execute(
            commands.AddEntry(
                resource_id=resource_id,
                user=user.identifier,
                message=data.message,
                entry=data.entry,
            )
        )
    except errors.IntegrityError as exc:
        return responses.JSONResponse(
            status_code=400,
            content={
                "error": str(exc),
                "errorCode": karp_errors.ClientErrorCodes.DB_INTEGRITY_ERROR,
            },
        )
    except errors.InvalidEntry as exc:
        return responses.JSONResponse(
            status_code=400,
            content={
                "error": str(exc),
                "errorCode": karp_errors.ClientErrorCodes.ENTRY_NOT_VALID,
            },
        )

    return {"newID": new_entry.entity_id}


@router.post("/{resource_id}/preview")
def preview_entry(
    resource_id: str,
    data: schemas.EntryAdd,
    user: auth.User = Security(deps.get_user_optional, scopes=["read"]),
    auth_service: auth.AuthService = Depends(deps.get_auth_service),
    preview_entry: search.PreviewEntry = Depends(
        deps.inject_from_req(search.PreviewEntry)
    ),
):
    if not auth_service.authorize(PermissionLevel.read, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="lexica:read"'},
        )
    try:
        input_dto = search.PreviewEntryInputDto(
            resource_id=resource_id, entry=data.entry, user=user.identifier
        )
    except pydantic.ValidationError as err:
        logger.exception("data is not valid")
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail={"error": str(err)}
        )
    else:
        return preview_entry.query(input_dto)


@router.post(
    "/{resource_id}/{entry_id}/update",
    tags=["Editing"],
    response_model=schemas.EntryAddResponse,
)
@router.post(
    "/{resource_id}/{entry_id}",
    tags=["Editing"],
    response_model=schemas.EntryAddResponse,
)
def update_entry(
    response: Response,
    resource_id: str,
    entry_id: UniqueIdStr,
    data: schemas.EntryUpdate,
    user: User = Security(deps.get_user, scopes=["write"]),
    auth_service: AuthService = Depends(deps.get_auth_service),
    updating_entry_uc: lex.UpdatingEntry = Depends(deps.get_lex_uc(lex.UpdatingEntry)),
):
    if not auth_service.authorize(PermissionLevel.write, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="write"'},
        )

    #     force_update = convert.str2bool(request.args.get("force", "false"))
    #     data = request.get_json()
    #     version = data.get("version")
    #     entry = data.get("entry")
    #     message = data.get("message")
    #     if not (version and entry and message):
    #         raise KarpError("Missing version, entry or message")
    logger.info(
        "updating entry",
        extra={
            "resource_id": resource_id,
            "entry_id": entry_id,
            "data": data,
            "user": user.identifier,
        },
    )
    try:
        entry = updating_entry_uc.execute(
            commands.UpdateEntry(
                resource_id=resource_id,
                entity_id=unique_id.parse(entry_id),
                version=data.version,
                user=user.identifier,
                message=data.message,
                entry=data.entry,
            )
        )
        # new_entry = entries.add_entry(
        #     resource_id, data.entry, user.identifier, message=data.message
        # )
        # new_id = entries.update_entry(
        #     resource_id,
        #     entry_id,
        #     data.version,
        #     data.entry,
        #     user.identifier,
        #     message=data.message,
        #     # force=force_update,
        # )
        return {"newID": entry.entity_id}
    except errors.EntryNotFound:
        return responses.JSONResponse(
            status_code=404,
            content={
                "error": f"Entry '{entry_id}' not found in resource '{resource_id}' (version=latest)",
                "errorCode": karp_errors.ClientErrorCodes.ENTRY_NOT_FOUND,
                "resource": resource_id,
                "entry_id": entry_id,
            },
        )
    except errors.UpdateConflict as err:
        err.error_obj["errorCode"] = karp_errors.ClientErrorCodes.VERSION_CONFLICT
        return responses.JSONResponse(
            status_code=status.HTTP_400_BAD_REQUEST, content=err.error_obj
        )
    except Exception as err:
        logger.exception(
            "error occured",
            extra={"resource_id": resource_id, "entry_id": entry_id, "data": data},
        )
        raise


@router.delete(
    "/{resource_id}/{entry_id}/delete",
    tags=["Editing"],
    status_code=status.HTTP_204_NO_CONTENT,
)
@router.delete(
    "/{resource_id}/{entry_id}",
    tags=["Editing"],
    status_code=status.HTTP_204_NO_CONTENT,
)
def delete_entry(
    resource_id: str,
    entry_id: UniqueIdStr,
    user: User = Security(deps.get_user, scopes=["write"]),
    auth_service: AuthService = Depends(deps.get_auth_service),
    deleting_entry_uc: lex.DeletingEntry = Depends(deps.get_lex_uc(lex.DeletingEntry)),
):
    """Delete a entry from a resource."""
    if not auth_service.authorize(PermissionLevel.write, user, [resource_id]):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not enough permissions",
            headers={"WWW-Authenticate": 'Bearer scope="lexica:write"'},
        )
    try:
        deleting_entry_uc.execute(
            commands.DeleteEntry(
                resource_id=resource_id,
                entity_id=unique_id.parse(entry_id),
                user=user.identifier,
            )
        )
    except errors.EntryNotFound:
        return responses.JSONResponse(
            status_code=404,
            content={
                "error": f"Entry '{entry_id}' not found in resource '{resource_id}' (version=latest)",
                "errorCode": karp_errors.ClientErrorCodes.ENTRY_NOT_FOUND,
                "resource": resource_id,
                "entity_id": entry_id,
            },
        )
    return


def init_app(app):
    app.include_router(router)



"""add entry_repos table

Revision ID: d2f8d1c19b65
Revises: 5bb29472d33f
Create Date: 2021-10-27 22:08:06.924305

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy_utils import UUIDType
from sqlalchemy_json import NestedMutableJson


# revision identifiers, used by Alembic.
revision = "d2f8d1c19b65"
down_revision = "5bb29472d33f"
branch_labels = None
depends_on = None


def create_entry_repos():
    op.create_table(
        "entry_repos",
        sa.Column("history_id", sa.Integer(), nullable=False),
        sa.Column("entity_id", UUIDType, nullable=False),
        sa.Column("type", sa.String(length=64), nullable=False),
        sa.Column("connection_str", sa.String(length=128)),
        sa.Column("name", sa.String(length=64), nullable=False),
        sa.Column("config", NestedMutableJson, nullable=False),
        sa.Column("last_modified", sa.Float(precision=53), nullable=False),
        sa.Column("last_modified_by", sa.String(100), nullable=False),
        sa.Column("message", sa.String(100), nullable=False),
        sa.Column("discarded", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("history_id"),
    )


def upgrade():
    create_entry_repos()


def downgrade():
    op.drop_table("entry_repos")



import pydantic

from karp.foundation import commands
from karp.utility import time


class Command(pydantic.BaseModel, commands.Command):
    timestamp: float = pydantic.Field(default_factory=time.utc_now)

    class Config:
        arbitrary_types_allowed = True



from karp.search_infrastructure.queries.generic_entries import GenericPreviewEntry
from .generic_resources import GenericResourceViews
from .generic_search_service import GenericSearchService
from .es_query import EsQuery
from .es6_search_service import Es6SearchService



import abc
from datetime import datetime
import logging
import re
from typing import Dict, List, Optional, Any, Tuple, Union

import elasticsearch
from elasticsearch import exceptions as es_exceptions

from karp.foundation.events import EventBus

from karp.lex.domain.entities import Entry
from karp.search.application.repositories import (
    Index,
    IndexEntry,
    IndexUnitOfWork,
)


logger = logging.getLogger("karp")

KARP_CONFIGINDEX = "karp_config"
KARP_CONFIGINDEX_TYPE = "configs"


class ConfigRepository(abc.ABC):
    pass


class Es6ConfigRepository(Index):
    def __init__(self, es: elasticsearch.Elasticsearch):
        self.es: elasticsearch.Elasticsearch = es
        self.ensure_config_index_exist()
        analyzed_fields, sortable_fields = self._init_field_mapping()
        self.analyzed_fields: Dict[str, List[str]] = analyzed_fields
        self.sortable_fields: Dict[str, Dict[str, List[str]]] = sortable_fields

    def ensure_config_index_exist(self) -> None:
        if not self.es.indices.exists(index=KARP_CONFIGINDEX):
            self.es.indices.create(
                index=KARP_CONFIGINDEX,
                body={
                    "settings": {
                        "number_of_shards": 1,
                        "number_of_replicas": 1,
                        "refresh_interval": -1,
                    },
                    "mappings": {
                        KARP_CONFIGINDEX_TYPE: {
                            "dynamic": False,
                            "properties": {"index_name": {"type": "text"}},
                        }
                    },
                },
            )
    @property
    def seen(self):
        return []

    def create_index(self, resource_id, config):
        print("creating es mapping ...")
        mapping = create_es6_mapping(config)

        properties = mapping["properties"]
        properties["freetext"] = {"type": "text"}
        disabled_property = {"enabled": False}
        properties["_entry_version"] = disabled_property
        properties["_last_modified"] = disabled_property
        properties["_last_modified_by"] = disabled_property

        body = {
            "settings": {
                "number_of_shards": 1,
                "number_of_replicas": 1,
                "refresh_interval": -1,
            },
            "mappings": {"entry": mapping},
        }

        date = datetime.now().strftime("%Y-%m-%d-%H%M%S%f")
        index_name = resource_id + "_" + date
        print(f"creating index '{index_name}' ...")
        result = self.es.indices.create(index=index_name, body=body)
        if "error" in result:
            print("failed to create index")
            raise RuntimeError("failed to create index")
        print("index created")
        self._set_index_name_for_resource(resource_id, index_name)
        return index_name

    def _set_index_name_for_resource(self, resource_id: str, index_name: str) -> str:
        self.es.index(
            index=KARP_CONFIGINDEX,
            id=resource_id,
            doc_type=KARP_CONFIGINDEX_TYPE,
            body={"index_name": index_name},
        )
        return index_name

    def _get_index_name_for_resource(self, resource_id: str) -> str:
        try:
            res = self.es.get(
                index=KARP_CONFIGINDEX, id=resource_id, doc_type=KARP_CONFIGINDEX_TYPE
            )
        except es_exceptions.NotFoundError as err:
            logger.debug(
                "didn't find index_name for resource '%s' details: %s", resource_id, err)
            return self._set_index_name_for_resource(resource_id, resource_id)
        return res["_source"]["index_name"]

    def publish_index(self, resource_id: str):
        if self.es.indices.exists_alias(name=resource_id):
            self.es.indices.delete_alias(name=resource_id, index="*")

        index_name = self._get_index_name_for_resource(resource_id)
        self.on_publish_resource(resource_id, index_name)
        print(f"publishing '{resource_id}' => '{index_name}'")
        self.es.indices.put_alias(name=resource_id, index=index_name)

    def add_entries(self, resource_id: str, entries: List[IndexEntry]):
        index_name = self._get_index_name_for_resource(resource_id)
        index_to_es = []
        for entry in entries:
            assert isinstance(entry, IndexEntry)
            # entry.update(metadata.to_dict())
            index_to_es.append(
                {
                    "_index": index_name,
                    "_id": entry.id,
                    "_type": "entry",
                    "_source": entry.entry,
                }
            )

        elasticsearch.helpers.bulk(self.es, index_to_es, refresh=True)

    def delete_entry(
        self,
        resource_id: str,
        *,
        entry_id: Optional[str] = None,
        entry: Optional[Entry] = None,
    ):
        if not entry and not entry_id:
            raise ValueError("Must give either 'entry' or 'entry_id'.")
        if entry:
            entry_id = entry.entry_id
        self.es.delete(
            index=resource_id,
            doc_type="entry",
            id=entry_id,
            refresh=True,
        )

    @staticmethod
    def get_analyzed_fields_from_mapping(
        properties: Dict[str, Dict[str, Dict[str, Any]]]
    ) -> List[str]:
        analyzed_fields = []

        for prop_name, prop_values in properties.items():
            if "properties" in prop_values:
                res = Es6Index.get_analyzed_fields_from_mapping(
                    prop_values["properties"]
                )
                analyzed_fields.extend(
                    [prop_name + "." + prop for prop in res])
            else:
                if prop_values["type"] == "text":
                    analyzed_fields.append(prop_name)
        return analyzed_fields

    def _init_field_mapping(
        self,
    ) -> Tuple[Dict[str, List[str]], Dict[str, Dict[str, List[str]]]]:
        """
        Create a field mapping based on the mappings of elasticsearch
        currently the only information we need is if a field is analyzed (i.e. text)
        or not.
        """

        field_mapping: Dict[str, List[str]] = {}
        sortable_fields = {}
        # Doesn't work for tests, can't find resource_definition
        # for resource in resourcemgr.get_available_resources():
        #     mapping = self.es.indices.get_mapping(index=resource.resource_id)
        #     field_mapping[resource.resource_id] = parse_mapping(
        #         next(iter(mapping.values()))['mappings']['entry']['properties']
        #     )
        aliases = self._get_all_aliases()
        mapping: Dict[
            str, Dict[str, Dict[str, Dict[str, Dict]]]
        ] = self.es.indices.get_mapping()
        # print(f"mapping = {mapping}")
        for (alias, index) in aliases:
            if (
                "mappings" in mapping[index]
                and "entry" in mapping[index]["mappings"]
                and "properties" in mapping[index]["mappings"]["entry"]
            ):
                field_mapping[alias] = Es6Index.get_analyzed_fields_from_mapping(
                    mapping[index]["mappings"]["entry"]["properties"]
                )
                sortable_fields[alias] = Es6Index.create_sortable_map_from_mapping(
                    mapping[index]["mappings"]["entry"]["properties"]
                )
        return field_mapping, sortable_fields

    def _get_index_mappings(
        self, index: Optional[str] = None
    ) -> Dict[str, Dict[str, Dict[str, Dict[str, Dict]]]]:
        kwargs = {"index": index} if index is not None else {}
        return self.es.indices.get_mapping(**kwargs)

    def _get_all_aliases(self) -> List[Tuple[str, str]]:
        """
        :return: a list of tuples (alias_name, index_name)
        """
        result = self.es.cat.aliases(h="alias,index")
        print(f"{result}")
        index_names = []
        for index_name in result.split("\n")[:-1]:
            print(f"index_name = {index_name}")
            if index_name[0] != ".":
                groups = re.search(r"([^ ]*) +(.*)", index_name).groups()
                alias = groups[0]
                index = groups[1]
                index_names.append((alias, index))
        return index_names

    def translate_sort_fields(
        self, resources: List[str], sort_values: List[str]
    ) -> List[Union[str, Dict[str, Dict[str, str]]]]:
        """Translate sort field to ES sort fields.

        Arguments:
            sort_values {List[str]} -- values to sort by

        Returns:
            List[str] -- values that ES can sort by.
        """
        translated_sort_fields: List[Union[str,
                                           Dict[str, Dict[str, str]]]] = []
        for sort_value in sort_values:
            sort_order = None
            if "|" in sort_value:
                sort_value, sort_order = sort_value.split("|", 1)
            for resource_id in resources:
                if sort_order:
                    translated_sort_fields.extend(
                        (
                            {field: {"order": sort_order}}
                            for field in self.translate_sort_field(
                                resource_id, sort_value
                            )
                        )
                    )
                translated_sort_fields.extend(
                    self.translate_sort_field(resource_id, sort_value)
                )

        return translated_sort_fields

    def translate_sort_field(self, resource_id: str, sort_value: str) -> List[str]:
        print(
            f"es6_indextranslate_sort_field: sortable_fields[{resource_id}] = {self.sortable_fields[resource_id]}"
        )
        if sort_value in self.sortable_fields[resource_id]:
            return self.sortable_fields[resource_id][sort_value]
        else:
            raise UnsupportedField(
                f"You can't sort by field '{sort_value}' for resource '{resource_id}'"
            )

    def on_publish_resource(self, alias_name: str, index_name: str):
        mapping = self._get_index_mappings(index=index_name)
        if (
            "mappings" in mapping[index_name]
            and "entry" in mapping[index_name]["mappings"]
            and "properties" in mapping[index_name]["mappings"]["entry"]
        ):
            self.analyzed_fields[
                alias_name
            ] = Es6Index.get_analyzed_fields_from_mapping(
                mapping[index_name]["mappings"]["entry"]["properties"]
            )
            self.sortable_fields[
                alias_name
            ] = Es6Index.create_sortable_map_from_mapping(
                mapping[index_name]["mappings"]["entry"]["properties"]
            )

    @staticmethod
    def create_sortable_map_from_mapping(properties: Dict) -> Dict[str, List[str]]:
        sortable_map = {}

        def parse_prop_value(sort_map, base_name, prop_name, prop_value: Dict):
            if "properties" in prop_value:
                for ext_name, ext_value in prop_value["properties"].items():
                    ext_base_name = f"{base_name}.{ext_name}"
                    parse_prop_value(sort_map, ext_base_name,
                                     ext_base_name, ext_value)
                return
            if prop_value["type"] in [
                "boolean",
                "date",
                "double",
                "keyword",
                "long",
                "ip",
            ]:
                sort_map[base_name] = [prop_name]
                sort_map[prop_name] = [prop_name]
                return
            if prop_value["type"] == "text":
                if "fields" in prop_value:
                    for ext_name, ext_value in prop_value["fields"].items():
                        parse_prop_value(
                            sort_map, base_name, f"{base_name}.{ext_name}", ext_value
                        )
                return

        for prop_name, prop_value in properties.items():
            parse_prop_value(sortable_map, prop_name, prop_name, prop_value)
            # if prop_value["type"] in ["boolean", "date", "double", "keyword", "long", "ip"]:
            #     sortable_map[prop_name] = prop_name
            # if prop_value["type"] == "text":
            #     if "fields" in prop_value:

        return sortable_map


# def parse_sortable_fields(properties: Dict[str, Any]) -> Dict[str, List[str]]:
#     for prop_name, prop_value in properties.items():
#         if prop_value["type"] in ["boolean", "date", "double", "keyword", "long", "ip"]:
#             return [prop_name]


def _create_es_mapping(config):
    es_mapping = {"dynamic": False, "properties": {}}

    fields = config["fields"]

    def recursive_field(parent_schema, parent_field_name, parent_field_def):
        if parent_field_def.get("virtual", False):
            fun = parent_field_def["function"]
            if list(fun.keys())[0] == "multi_ref":
                res_object = fun["multi_ref"]["result"]
                recursive_field(parent_schema, "v_" +
                                parent_field_name, res_object)
            if "result" in fun:
                res_object = fun["result"]
                recursive_field(parent_schema, "v_" +
                                parent_field_name, res_object)
            return
        if parent_field_def.get("ref"):
            if "field" in parent_field_def["ref"]:
                res_object = parent_field_def["ref"]["field"]
            else:
                res_object = {}
                res_object.update(parent_field_def)
                del res_object["ref"]
            recursive_field(parent_schema, "v_" +
                            parent_field_name, res_object)
        if parent_field_def["type"] != "object":
            # TODO this will not work when we have user defined types, s.a. saldoid
            # TODO number can be float/non-float, strings can be keyword or text in need of analyzing etc.
            if parent_field_def["type"] == "integer":
                mapped_type = "long"
            elif parent_field_def["type"] == "number":
                mapped_type = "double"
            elif parent_field_def["type"] == "boolean":
                mapped_type = "boolean"
            elif parent_field_def["type"] == "string":
                mapped_type = "text"
            elif parent_field_def['type'] == 'long_string':
                mapped_type = 'text'
            else:
                mapped_type = "keyword"
            result = {"type": mapped_type}
            if parent_field_def['type'] == 'string':
                result["fields"] = {"raw": {"type": "keyword"}}
        else:
            result = {"properties": {}}

            for child_field_name, child_field_def in parent_field_def["fields"].items():
                recursive_field(result, child_field_name, child_field_def)

        parent_schema["properties"][parent_field_name] = result

    for field_name, field_def in fields.items():
        print(f"creating mapping for field '{field_name}'")
        recursive_field(es_mapping, field_name, field_def)

    return es_mapping


def create_es6_mapping(config: Dict) -> Dict:
    mapping = _create_es_mapping(config)
    mapping['settings'] = {
        'analysis': {
            'analyzer': {
                'default': {
                    'char_filter': [
                        'compound',
                        'swedish_aa',
                        'swedish_ae',
                        'swedish_oe'
                    ],
                    'filter': [
                        'swedish_folding',
                        'lowercase'
                    ],
                    'tokenizer': 'standard'
                }
            },
            'char_filter': {
                'compound': {
                    'pattern': '-',
                    'replacement': '',
                    'type': 'pattern_replace'
                },
                'swedish_aa': {
                    'pattern': '[Ǻǻ]',
                    'replacement': 'å',
                    'type': 'pattern_replace'
                },
                'swedish_ae': {
                    'pattern': '[æÆǞǟ]',
                    'replacement': 'ä',
                    'type': 'pattern_replace'
                },
                'swedish_oe': {
                    'pattern': '[ØøŒœØ̈ø̈ȪȫŐőÕõṌṍṎṏȬȭǾǿǬǭŌōṒṓṐṑ]',
                    'replacement': 'ö',
                    'type': 'pattern_replace'
                },
            },
            'filter': {
                'swedish_folding': {
                    'type': 'icu_folding',
                    'unicodeSetFilter': '[^åäöÅÄÖ]'
                },
                'swedish_sort': {
                    'language': 'sv',
                    'type': 'icu_collation'
                }
            }
        }
    }
    return mapping


class Es6IndexUnitOfWork(
    IndexUnitOfWork
):
    def __init__(
        self,
        es: elasticsearch.Elasticsearch,
        event_bus: EventBus,
    ) -> None:
        super().__init__(event_bus=event_bus)
        self._index = Es6Index(es=es)

    # @classmethod
    # def from_dict(cls, **kwargs):
    #     return cls()

    def _commit(self):
        logger.debug('Calling _commit in Es6IndexUnitOfWork')

    def rollback(self):
        return super().rollback()

    @property
    def repo(self) -> Es6Index:
        return self._index

    def _close(self):
        pass
import json
import logging
import re
from datetime import datetime
from typing import Any, Dict, List, Optional, Set, Tuple, Union

import elasticsearch
import elasticsearch.helpers  # pyre-ignore
import elasticsearch_dsl as es_dsl  # pyre-ignore

# from karp import query_dsl
from karp.search.application.queries import (
    QueryRequest,
    SearchService,
)
from karp.search.domain.errors import \
    UnsupportedField  # IncompleteQuery,; UnsupportedQuery,
from karp.lex.domain.entities.entry import Entry
from karp.lex.domain.entities.resource import Resource
from karp.search.domain import query_dsl
from karp.search_infrastructure.elasticsearch6 import es_config
from .es_query import EsQuery

# from karp.query_dsl import basic_ast as ast, op, is_a


logger = logging.getLogger("karp")

KARP_CONFIGINDEX = "karp_config"
KARP_CONFIGINDEX_TYPE = "configs"


class EsQueryBuilder(query_dsl.NodeWalker):
    def walk_object(self, node):
        return node

    def walk__and(self, node):
        result = self.walk(node.exps[0])
        for n in node.exps[1:]:
            result = result & self.walk(n)

        return result

    def walk__contains(self, node):
        return es_dsl.Q(
            "regexp", **{self.walk(node.field): f".*{self.walk(node.arg)}.*"}
        )

    def walk__endswith(self, node):
        return es_dsl.Q("regexp", **{self.walk(node.field): f".*{self.walk(node.arg)}"})

    # def walk__equals(self, node):
    #     return es_dsl.Q(
    #         "match",
    #         **{
    #             self.walk(node.field): {"query": self.walk(node.arg), "operator": "and"}
    #         },
    #     )

    def walk__equals(self, node):
        return es_dsl.Q(
            "match",
            **{self.walk(node.field): {"query": self.walk(node.arg), "operator": "and"}}
        )

    def walk__exists(self, node):
        return es_dsl.Q("exists", field=self.walk(node.field))

    def walk__freergxp(self, node):
        return es_dsl.Q(
            "query_string", query=f"/{self.walk(node.arg)}/", default_field="*"
        )

    def walk__freetext_string(self, node):
        return es_dsl.Q("multi_match", query=self.walk(node.arg), fuzziness=1)

    def walk__freetext_any_but_string(self, node):
        return es_dsl.Q("multi_match", query=self.walk(node.arg))

    def walk_range(self, node):
        return es_dsl.Q(
            "range",
            **{self.walk(node.field): {self.walk(node.op): self.walk(node.arg)}},
        )

    walk__gt = walk_range
    walk__gte = walk_range
    walk__lt = walk_range
    walk__lte = walk_range

    def walk__missing(self, node):
        return es_dsl.Q(
            "bool", must_not=es_dsl.Q("exists", field=self.walk(node.field))
        )

    def walk__not(self, node):
        return ~self.walk(node.expr)

    def walk__or(self, node):
        result = self.walk(node.exps[0])
        for n in node.exps[1:]:
            result = result | self.walk(n)

        return result

    def walk__regexp(self, node):
        return es_dsl.Q("regexp", **{self.walk(node.field): self.walk(node.arg)})

    def walk__startswith(self, node):
        return es_dsl.Q("regexp", **{self.walk(node.field): f"{self.walk(node.arg)}.*"})


class Es6SearchService(SearchService):
    def __init__(self, es: elasticsearch.Elasticsearch):
        self.es: elasticsearch.Elasticsearch = es
        self.query_builder = EsQueryBuilder()
        self.parser = query_dsl.KarpQueryV6Parser(
            semantics=query_dsl.KarpQueryV6ModelBuilderSemantics())
        if not self.es.indices.exists(index=KARP_CONFIGINDEX):
            self.es.indices.create(
                index=KARP_CONFIGINDEX,
                body={
                    "settings": {
                        "number_of_shards": 1,
                        "number_of_replicas": 1,
                        "refresh_interval": -1,
                    },
                    "mappings": {
                        KARP_CONFIGINDEX_TYPE: {
                            "dynamic": False,
                            "properties": {"index_name": {"type": "text"}},
                        }
                    },
                },
            )
        analyzed_fields, sortable_fields = self._init_field_mapping()
        self.analyzed_fields: Dict[str, List[str]] = analyzed_fields
        self.sortable_fields: Dict[str, Dict[str, List[str]]] = sortable_fields

    def _get_index_name_for_resource(self, resource_id: str) -> str:
        res = self.es.get(
            index=KARP_CONFIGINDEX, id=resource_id, doc_type=KARP_CONFIGINDEX_TYPE
        )
        return res["_source"]["index_name"]

    @staticmethod
    def get_analyzed_fields_from_mapping(
        properties: Dict[str, Dict[str, Dict[str, Any]]]
    ) -> List[str]:
        analyzed_fields = []

        for prop_name, prop_values in properties.items():
            if "properties" in prop_values:
                res = Es6SearchService.get_analyzed_fields_from_mapping(
                    prop_values["properties"]
                )
                analyzed_fields.extend(
                    [prop_name + "." + prop for prop in res])
            else:
                if prop_values["type"] == "text":
                    analyzed_fields.append(prop_name)
        return analyzed_fields

    def _init_field_mapping(
        self,
    ) -> Tuple[Dict[str, List[str]], Dict[str, Dict[str, List[str]]]]:
        """
        Create a field mapping based on the mappings of elasticsearch
        currently the only information we need is if a field is analyzed (i.e. text)
        or not.
        """

        field_mapping: Dict[str, List[str]] = {}
        sortable_fields = {}
        # Doesn't work for tests, can't find resource_definition
        # for resource in resourcemgr.get_available_resources():
        #     mapping = self.es.indices.get_mapping(index=resource.resource_id)
        #     field_mapping[resource.resource_id] = parse_mapping(
        #         next(iter(mapping.values()))['mappings']['entry']['properties']
        #     )
        aliases = self._get_all_aliases()
        mapping: Dict[
            str, Dict[str, Dict[str, Dict[str, Dict]]]
        ] = self.es.indices.get_mapping()
        # print(f"mapping = {mapping}")
        for (alias, index) in aliases:
            if (
                "mappings" in mapping[index]
                and "entry" in mapping[index]["mappings"]
                and "properties" in mapping[index]["mappings"]["entry"]
            ):
                field_mapping[alias] = Es6SearchService.get_analyzed_fields_from_mapping(
                    mapping[index]["mappings"]["entry"]["properties"]
                )
                sortable_fields[alias] = Es6SearchService.create_sortable_map_from_mapping(
                    mapping[index]["mappings"]["entry"]["properties"]
                )
        return field_mapping, sortable_fields

    def _get_index_mappings(
        self, index: Optional[str] = None
    ) -> Dict[str, Dict[str, Dict[str, Dict[str, Dict]]]]:
        kwargs = {"index": index} if index is not None else {}
        return self.es.indices.get_mapping(**kwargs)

    def _get_all_aliases(self) -> List[Tuple[str, str]]:
        """
        :return: a list of tuples (alias_name, index_name)
        """
        result = self.es.cat.aliases(h="alias,index")
        print(f"{result}")
        index_names = []
        for index_name in result.split("\n")[:-1]:
            print(f"index_name = {index_name}")
            if index_name[0] != ".":
                groups = re.search(r"([^ ]*) +(.*)", index_name).groups()
                alias = groups[0]
                index = groups[1]
                index_names.append((alias, index))
        return index_names

    def build_query(self, args, resource_str: str) -> EsQuery:
        query = EsQuery()
        query.parse_arguments(args, resource_str)
        return query

    def _format_result(self, resource_ids, response):
        logger.debug(
            "es6_search_service_format_result called with resource_ids=%s", resource_ids
        )

        def format_entry(entry):

            dict_entry = entry.to_dict()
            version = dict_entry.pop("_entry_version", None)
            last_modified_by = dict_entry.pop("_last_modified_by", None)
            last_modified = dict_entry.pop("_last_modified", None)
            return {
                "id": entry.meta.id,
                "version": version,
                "last_modified": last_modified,
                "last_modified_by": last_modified_by,
                "resource": next(
                    resource
                    for resource in resource_ids
                    if entry.meta.index.startswith(resource)
                ),
                "entry": dict_entry,
            }

        result = {
            "total": response.hits.total,
            "hits": [format_entry(entry) for entry in response],
        }
        return result

    def query(self, request: QueryRequest):
        print(f"query called with {request}")
        query = EsQuery.from_query_request(request)
        return self.search_with_query(query)

    def query_split(self, request: QueryRequest):
        print(f"query called with {request}")
        query = EsQuery.from_query_request(request)
        query.split_results = True
        return self.search_with_query(query)

    def search_with_query(self, query: EsQuery):
        logger.info("search_with_query called with query=%s", query)
        print("search_with_query called with query={}".format(query))
        es_query = None
        if query.q:
            model = self.parser(query.q)
            es_query = self.query_builder.walk(model)
        if query.split_results:
            ms = es_dsl.MultiSearch(using=self.es)

            for resource in query.resources:
                s = es_dsl.Search(index=resource)

                if es_query is not None:
                    s = s.query(es_query)
                s = s[query.from_: query.from_ + query.size]
                if query.sort:
                    s = s.sort(
                        *self.translate_sort_fields([resource], query.sort))
                elif resource in query.sort_dict:
                    s = s.sort(
                        *self.translate_sort_fields(
                            [resource], query.sort_dict[resource]
                        )
                    )
                ms = ms.add(s)

            responses = ms.execute()
            result = {"total": 0, "hits": {}}
            for i, response in enumerate(responses):
                result["hits"][query.resources[i]] = self._format_result(
                    query.resources, response
                ).get("hits", [])
                result["total"] += response.hits.total
                if query.lexicon_stats:
                    if "distribution" not in result:
                        result["distribution"] = {}
                    result["distribution"][query.resources[i]
                                           ] = response.hits.total
            return result
        else:
            s = es_dsl.Search(
                using=self.es, index=query.resources, doc_type="entry")
            if es_query is not None:
                s = s.query(es_query)

            s = s[query.from_: query.from_ + query.size]

            if query.lexicon_stats:
                s.aggs.bucket(
                    "distribution", "terms", field="_index", size=len(query.resources)
                )
            if query.sort:
                s = s.sort(
                    *self.translate_sort_fields(query.resources, query.sort))
            elif query.sort_dict:
                sort_fields = []
                for resource, sort in query.sort_dict.items():
                    sort_fields.extend(
                        self.translate_sort_fields([resource], sort))
                s = s.sort(*sort_fields)
            logger.debug("s = %s", s.to_dict())
            response = s.execute()

            # TODO format response in a better way, because the whole response takes up too much space in the logs
            # logger.debug('response = {}'.format(response.to_dict()))

            logger.debug("calling _format_result")
            result = self._format_result(query.resources, response)
            if query.lexicon_stats:
                result["distribution"] = {}
                for bucket in response.aggregations.distribution.buckets:
                    key = bucket["key"]
                    value = bucket["doc_count"]
                    result["distribution"][key.rsplit("_", 1)[0]] = value

            # logger.debug("return result = %s", result)
            return result

    def translate_sort_fields(
        self, resources: List[str], sort_values: List[str]
    ) -> List[Union[str, Dict[str, Dict[str, str]]]]:
        """Translate sort field to ES sort fields.

        Arguments:
            sort_values {List[str]} -- values to sort by

        Returns:
            List[str] -- values that ES can sort by.
        """
        translated_sort_fields: List[Union[str,
                                           Dict[str, Dict[str, str]]]] = []
        for sort_value in sort_values:
            sort_order = None
            if "|" in sort_value:
                sort_value, sort_order = sort_value.split("|", 1)
            for resource_id in resources:
                if sort_order:
                    translated_sort_fields.extend(
                        (
                            {field: {"order": sort_order}}
                            for field in self.translate_sort_field(
                                resource_id, sort_value
                            )
                        )
                    )
                translated_sort_fields.extend(
                    self.translate_sort_field(resource_id, sort_value)
                )

        return translated_sort_fields

    def translate_sort_field(self, resource_id: str, sort_value: str) -> List[str]:
        print(
            f"es6_search_servicetranslate_sort_field: sortable_fields[{resource_id}] = {self.sortable_fields[resource_id]}"
        )
        if sort_value in self.sortable_fields[resource_id]:
            return self.sortable_fields[resource_id][sort_value]
        else:
            raise UnsupportedField(
                f"You can't sort by field '{sort_value}' for resource '{resource_id}'"
            )

    def search_ids(self, resource_id: str, entry_ids: str):
        logger.info(
            "Called EsSearch.search_ids(self, args, resource_id, entry_ids) with:"
        )
        logger.info("  resource_id = {}".format(resource_id))
        logger.info("  entry_ids = {}".format(entry_ids))
        entries = entry_ids.split(",")
        query = es_dsl.Q("terms", _id=entries)
        logger.debug("query = {}".format(query))
        s = es_dsl.Search(using=self.es, index=resource_id).query(query)
        logger.debug("s = {}".format(s.to_dict()))
        response = s.execute()

        return self._format_result([resource_id], response)

    def statistics(self, resource_id: str, field: str):
        s = es_dsl.Search(using=self.es, index=resource_id)
        s = s[0:0]

        if field in self.analyzed_fields[resource_id]:
            field += ".raw"

        logger.debug("Statistics: analyzed fields are:")
        logger.debug(json.dumps(self.analyzed_fields, indent=4))
        logger.debug(
            "Doing aggregations on resource_id: {resource_id}, on field {field}".format(
                resource_id=resource_id, field=field
            )
        )
        s.aggs.bucket("field_values", "terms", field=field, size=2147483647)
        response = s.execute()
        print(f'{response=}')
        return [
            {"value": bucket["key"], "count": bucket["doc_count"]}
            for bucket in response.aggregations.field_values.buckets
        ]

    def on_publish_resource(self, alias_name: str, index_name: str):
        mapping = self._get_index_mappings(index=index_name)
        if (
            "mappings" in mapping[index_name]
            and "entry" in mapping[index_name]["mappings"]
            and "properties" in mapping[index_name]["mappings"]["entry"]
        ):
            self.analyzed_fields[
                alias_name
            ] = Es6SearchService.get_analyzed_fields_from_mapping(
                mapping[index_name]["mappings"]["entry"]["properties"]
            )
            self.sortable_fields[
                alias_name
            ] = Es6SearchService.create_sortable_map_from_mapping(
                mapping[index_name]["mappings"]["entry"]["properties"]
            )

    @staticmethod
    def create_sortable_map_from_mapping(properties: Dict) -> Dict[str, List[str]]:
        sortable_map = {}

        def parse_prop_value(sort_map, base_name, prop_name, prop_value: Dict):
            if "properties" in prop_value:
                for ext_name, ext_value in prop_value["properties"].items():
                    ext_base_name = f"{base_name}.{ext_name}"
                    parse_prop_value(sort_map, ext_base_name,
                                     ext_base_name, ext_value)
                return
            if prop_value["type"] in [
                "boolean",
                "date",
                "double",
                "keyword",
                "long",
                "ip",
            ]:
                sort_map[base_name] = [prop_name]
                sort_map[prop_name] = [prop_name]
                return
            if prop_value["type"] == "text":
                if "fields" in prop_value:
                    for ext_name, ext_value in prop_value["fields"].items():
                        parse_prop_value(
                            sort_map, base_name, f"{base_name}.{ext_name}", ext_value
                        )
                return

        for prop_name, prop_value in properties.items():
            parse_prop_value(sortable_map, prop_name, prop_name, prop_value)
            # if prop_value["type"] in ["boolean", "date", "double", "keyword", "long", "ip"]:
            #     sortable_map[prop_name] = prop_name
            # if prop_value["type"] == "text":
            #     if "fields" in prop_value:

        return sortable_map


# def parse_sortable_fields(properties: Dict[str, Any]) -> Dict[str, List[str]]:
#     for prop_name, prop_value in properties.items():
#         if prop_value["type"] in ["boolean", "date", "double", "keyword", "long", "ip"]:
#             return [prop_name]






import logging
from typing import List

from karp.foundation import value_objects
from karp.auth.domain.entities.user import User
from karp.auth.domain import auth_service


logger = logging.getLogger(__name__)

class DummyAuthService(auth_service.AuthService):
    def __init__(self):
        logger.warning("Using DummyAuthService: Don't use this in production!")

    def authenticate(self, scheme: str, credentials: str) -> User:
        return User(credentials or "dummy", {}, {})

    def authorize(
        self,
        level: value_objects.PermissionLevel,
        user: User,
        resource_ids: List[str],
    ) -> bool:
        return True



import abc
import enum
import typing

import pydantic

from karp.foundation.value_objects.permission_level import PermissionLevel


class Scope(str, enum.Enum):
    admin = 'ADMIN'
    write = 'WRITE'
    read = 'READ'


class ResourcePermissionDto(pydantic.BaseModel):
    resource_id: str
    protected: typing.Optional[Scope]


class GetResourcePermissions(abc.ABC):
    @abc.abstractmethod
    def query(self) -> typing.List[ResourcePermissionDto]:
        pass


class IsResourceProtected(abc.ABC):
    @abc.abstractmethod
    def query(self, resource_id: str, level: PermissionLevel) -> bool:
        return True





from karp.search_infrastructure.repositories.es6_indicies import Es6Index, Es6IndexUnitOfWork


class TestEs6Index:
    def test_can_instantiate_es6_index(self):
        try:
            Es6Index(None)
        except AttributeError:
            pass

    def test_can_instantiate_es6_index_uow(self):
        try:
            Es6IndexUnitOfWork(None, event_bus=None)
        except AttributeError:
            pass



import typing
from enum import Enum
from typing import Dict, Optional

from pydantic import BaseModel

from karp.foundation.value_objects import unique_id


class SystemResponse(BaseModel):
    message: str = "ok"

    def __bool__(self):
        return True


class SystemOk(SystemResponse):
    pass


class SystemNotOk(SystemResponse):
    def __bool__(self):
        return False


class SystemMonitorResponse(BaseModel):
    database: str


class PermissionLevel(str, Enum):
    write = "write"
    read = "read"
    admin = "admin"


class EntryFormat(str, Enum):
    json = "json"
    csv = "csv"
    xml = "xml"
    lmf = "lmf?"
    tsb = "tsb"


class Entry(BaseModel):
    entry_id: str
    resource: str
    version: int
    entry: typing.Dict
    last_modified_by: str
    last_modified: float


class EntryBase(BaseModel):
    entry: Dict


class EntryAdd(EntryBase):
    message: str = ""


class EntryUpdate(EntryBase):
    message: str
    version: int


class EntityIdMixin(BaseModel):
    entity_id: unique_id.UniqueIdStr


class EntryAddResponse(BaseModel):
    newID: unique_id.UniqueIdStr


class ResourceBase(BaseModel):
    resource_id: str
    name: str
    config: typing.Dict
    message: Optional[str] = None
    entry_repo_id: Optional[unique_id.UniqueIdStr] = None
    is_published: Optional[bool] = None
    version: Optional[int] = None


class ResourceCreate(ResourceBase):
    pass


class ResourcePublic(EntityIdMixin, ResourceBase):
    last_modified: float


class ResourceProtected(ResourcePublic):
    last_modified_by: str


class ResourcePublish(BaseModel):
    message: str
    version: int
    resource_id: Optional[str]



import os

import environs
from sqlalchemy.engine import URL as DatabaseUrl, make_url
from starlette.config import Config
from starlette.datastructures import Secret

PROJECT_NAME = 'Karp'
VERSION = '6.1.0'
API_PREFIX = '/'
# SECRET_KEY = config("SECRET_KEY", cast=Secret, default="CHANGEME")


def load_env() -> environs.Env:
    config_path = os.environ.get("CONFIG_PATH", ".env")
    env = environs.Env()
    env.read_env(config_path)
    return env


def parse_database_name(env: environs.Env) -> str:
    database_name = env('DB_DATABASE', 'karp')
    if env('TESTING', None):
        database_name = env('DB_TEST_DATABASE',
                            None) or f'{database_name}_test'
    return database_name


def parse_database_url(env: environs.Env) -> DatabaseUrl:
    database_url = env('DATABASE_URL', None)
    if env.bool('TESTING', False):
        database_test_url = env('DATABASE_TEST_URL', None)
        if database_test_url:
            return make_url(database_test_url)
        elif database_url:
            return make_url(f'{database_url}_test')

    if database_url:
        return make_url(database_url)

    database_name = parse_database_name(env)

    return DatabaseUrl.create(
        drivername=env('DB_DRIVER', 'mysql+pymysql'),
        username=env('DB_USER', None),
        password=env('DB_PASSWORD', None),
        host=env('DB_HOST', None),
        port=env.int('DB_PORT', None),
        database=database_name,
        query={'charset': 'utf8mb4'}
    )


def parse_database_url_wo_db(env: environs.Env) -> DatabaseUrl:
    return DatabaseUrl.create(
        drivername=env('DB_DRIVER', 'mysql+pymysql'),
        username=env('DB_USER', None),
        password=env('DB_PASSWORD', None),
        host=env('DB_HOST', None),
        port=env.int('DB_PORT', None),
        query={'charset': 'utf8mb4'}
    )


config = load_env()


DATABASE_URL = parse_database_url(config)
DATABASE_URL_WO_DB = parse_database_url_wo_db(config)
DATABASE_NAME = parse_database_name(config)

AUTH_JWT_AUDIENCE = 'spraakbanken:auth'


def parse_sqlalchemy_url_wo_db(env: environs.Env) -> DatabaseUrl:
    return DatabaseUrl.create(
        drivername=env('DB_DRIVER', 'mysql+pymysql'),
        username=env('DB_USER', None),
        password=env('DB_PASSWORD', None),
        host=env('DB_HOST', None),
        port=env.int('DB_PORT', None),
        database='',
        query={'charset': 'utf8mb4'}
    )



"""
Perform health checks on the server.

Used to perform readiness and liveness probes on the server.
"""

from fastapi import APIRouter, Depends, Response, status

from karp.webapp import schemas


router = APIRouter(tags=["Health"])



@router.get(
    "/",
    response_model=schemas.SystemMonitorResponse,
    include_in_schema=False
)
def perform_health_check(
):
    return {"database": "ok"}


def init_app(app):
    app.include_router(router)



import enum
from typing import IO, Optional

NoIndexModuleConfigured = 10


class ClientErrorCodes(enum.IntEnum):
    UNKNOWN_ERROR = 1
    RESOURCE_DOES_NOT_EXIST = 20
    RESOURCE_NOT_PUBLISHED = 21
    RESOURCE_CONFIG_NOT_VALID = 22
    RESOURCE_CONFIG_CANNOT_UPDATE = 23
    RESOURCE_ALREADY_PUBLISHED = 24
    ENTRY_NOT_FOUND = 30
    ENTRY_NOT_CHANGED = 31
    ENTRY_NOT_VALID = 32
    VERSION_CONFLICT = 33
    ENTRY_ID_MISMATCH = 34
    EXPIRED_JWT = 40
    NOT_PERMITTED = 41
    AUTH_GENERAL_ERROR = 49
    BAD_PARAMETER_FORMAT = 50
    DB_GENERAL_ERROR = 60
    DB_INTEGRITY_ERROR = 61
    PLUGIN_DOES_NOT_EXIT = 70
    SEARCH_GENERAL_ERROR = 80
    SEARCH_INCOMPLETE_QUERY = 81
    SEARCH_UNSUPPORTED_QUERY = 82
    SEARCH_UNSUPPORTED_FIELD = 83


class KarpError(Exception):
    def __init__(self, message: str, code: int = None, http_return_code: int = 400):
        super().__init__(message)
        self.message = message
        self.code = code or ClientErrorCodes.UNKNOWN_ERROR
        self.http_return_code = http_return_code


class UserError(KarpError):
    pass


class ResourceNotFoundError(KarpError):
    def __init__(self, resource_id, version: int = None):
        super().__init__(
            "Resource not found. ID: {resource_id}, version: {version}".format(
                resource_id=resource_id, version=version
            ),
            ClientErrorCodes.RESOURCE_DOES_NOT_EXIST,
        )


class ResourceInvalidConfigError(KarpError):
    def __init__(self, resource_id, config_file: IO, validation_error_msg: str):
        msg_fmt = """
        Resource config is not valid.
        ID: {resource_id},
        config: {config_file},
        error: "{validation_error_msg}"
        """
        super().__init__(
            msg_fmt.format(
                resource_id=resource_id,
                config_file=config_file.name,
                validation_error_msg=validation_error_msg,
            ),
            ClientErrorCodes.RESOURCE_CONFIG_NOT_VALID,
        )


class ResourceConfigUpdateError(KarpError):
    def __init__(self, msg, resource_id, config_file: IO):
        msg_fmt = """
        Cannot update config for resource.
        Message: '{msg}'
        ID: {resource_id},
        config: {config_file}"
        """
        super().__init__(
            msg_fmt.format(
                resource_id=resource_id,
                config_file=config_file.name,
                msg=msg,
            ),
            ClientErrorCodes.RESOURCE_CONFIG_CANNOT_UPDATE,
        )


class ResourceAlreadyPublished(KarpError):
    def __init__(self, resource_id: str):
        super().__init__(
            f"Resource '{resource_id}' already published",
            code=ClientErrorCodes.RESOURCE_ALREADY_PUBLISHED,
        )


class EntryNotFoundError(KarpError):
    def __init__(
        self, resource_id, entry_id, entry_version=None, resource_version=None
    ):
        msg = "Entry '{entry_id}' (version {entry_version}) not found. resource_id: {resource_id}, version: {resource_version}"
        super().__init__(
            msg.format(
                resource_id=resource_id,
                resource_version=resource_version if resource_version else "latest",
                entry_id=entry_id,
                entry_version=entry_version if entry_version else "latest",
            ),
            ClientErrorCodes.ENTRY_NOT_FOUND,
        )


class UpdateConflict(KarpError):
    def __init__(self, diff):
        super().__init__(
            "Version conflict. Please update entry.",
            code=ClientErrorCodes.VERSION_CONFLICT,
        )
        self.error_obj = {"diff": diff,
                          "errorCode": self.code, "error": self.message}


class EntryIdMismatch(UserError):
    def __init__(self, new_entry_id: str, entry_id: str):
        super().__init__(
            f"entry_id '{new_entry_id}' does not equal '{entry_id}'",
            code=ClientErrorCodes.ENTRY_ID_MISMATCH,
        )


class PluginNotFoundError(KarpError):
    def __init__(self, plugin_id: str, resource_id: str = None):
        super().__init__(
            "Plugin '{plugin_id}' not found, referenced by '{resource_id}'".format(
                plugin_id=plugin_id, resource_id=resource_id if resource_id else "..."
            ),
            ClientErrorCodes.PLUGIN_DOES_NOT_EXIT,
        )


class IntegrityError(UserError):
    def __init__(
        self,
        message: Optional[str] = None,
        *,
        key: Optional[str] = None,
        value: Optional[str] = None,
    ) -> None:
        if message is None:
            message = ""
        if key and value:
            message = f"The key '{key}' is not unique (value='{value}')."
        super().__init__(
            message,
            code=ClientErrorCodes.DB_INTEGRITY_ERROR,
        )



import pytest

from typer import Typer
from typer.testing import CliRunner

from karp.main import config


class TestCli:
    def test_version(self, runner: CliRunner, cliapp: Typer):
        result = runner.invoke(cliapp, ['--version'])
        assert result.exit_code == 0
        assert f'{config.PROJECT_NAME} CLI {config.VERSION}\n' == result.stdout



import logging
from typing import Optional
import sys

from sqlalchemy.engine import Connection
from sqlalchemy.orm import Session
import typer

from karp.main import bootstrap_app, modules, config
from karp.cliapp import subapps

logger = logging.getLogger(__name__)


def create_app():
    app = typer.Typer(help="Karp CLI")
    app_context = bootstrap_app()

    @app.callback()
    def set_app_context(
        ctx: typer.Context,
        version: Optional[bool] = typer.Option(
            None, "--version", callback=version_callback, is_eager=True
        )
    ):
        if ctx.invoked_subcommand is None:
            ctx.obj = {}
        else:
            ctx.obj = {}
            ctx.obj['connection'] = app_context.container.get(Connection)
            ctx.obj['session'] = Session(bind=ctx.obj['connection'])
            logger.debug('create session', extra={
                         'session': ctx.obj['session']})
            ctx.obj['container'] = app_context.container.create_child_injector(
                modules.request_configuration(
                    conn=ctx.obj['connection'],
                    session=ctx.obj['session'],
                )
            )

    subapps.add_subapps(app)
    load_commands(app)

    return app


def version_callback(value: bool):
    if value:
        typer.echo(f'{config.PROJECT_NAME} CLI {config.VERSION}')
        raise typer.Exit()


def load_commands(app: typer.Typer):
    modules.load_modules('karp.clicommands', app=app)


cliapp = create_app()


if __name__ == "__main__":
    # cliapp = create_app()
    cliapp()



import pytest

from karp.lex.application import repositories
from karp.foundation.value_objects import unique_id
from . import adapters


def test_create_fake_entry_repo_uow(
    lex_ctx: adapters.UnitTestContext,
):
    uow_factory = lex_ctx.container.get(
        repositories.EntryRepositoryUnitOfWorkFactory)

    entry_repo_uow = uow_factory.create(
        "fake", unique_id.make_unique_id(), "name", {}, connection_str=None, user='kristoff@example.com', message='msg', timestamp=123456789)

    print(f"entry_repo_uow = {entry_repo_uow}")
    assert isinstance(entry_repo_uow, adapters.InMemoryEntryUnitOfWork)


@pytest.mark.skip()
def test_create_fake_entry_repo_uow2(
    lex_ctx: adapters.UnitTestContext,
):
    uow_factory = lex_ctx.container.get(
        repositories.EntryRepositoryUnitOfWorkFactory)

    entry_repo_uow = uow_factory.create(
        "fake2", unique_id.make_unique_id(), "name", {})

    print(f"entry_repo_uow = {entry_repo_uow}")
    assert isinstance(entry_repo_uow, adapters.InMemoryEntryUnitOfWork2)



from typing import Dict, Optional

import pytest

from karp.search_infrastructure.repositories.es6_indicies import create_es6_mapping


class TestCreateEs6Mapping:
    def test_minimal_valid_input(self):
        data = {
            'fields': {}
        }
        mapping = create_es6_mapping(data)

        assert mapping['properties'] == {}

    @pytest.mark.parametrize(
        'field, field_def, expected_property',
        [
            ('name', {'type': 'boolean'}, {}),
            ('name', {'type': 'number'}, {'type': 'double'}),
            ('name', {'type': 'integer'}, {'type': 'long'}),
            (
                'name',
                {'type': 'string'},
                {
                    'type': 'text',
                    'fields': {'raw': {'type': 'keyword'}}
                }
            ),
            ('name', {'type': 'long_string'}, {'type': 'text'}),
        ]
    )
    def test_standard_types(
        self,
        field: str,
        field_def: Dict,
        expected_property: Optional[Dict]
    ):
        data = {
            'fields': {
                field: field_def,
            },
        }
        mapping = create_es6_mapping(data)
        assert field in mapping['properties']
        expected_property = expected_property or field_def
        assert mapping['properties'][field] == expected_property

    @pytest.mark.parametrize(
        'field, field_def, expected_property',
        [
            (
                'name',
                {
                    'type': 'object',
                    'fields': {
                        'first': {
                            'type': 'string'
                        }
                    }
                },
                {
                    'properties': {
                        'first': {
                            'type': 'text',
                            'fields': {
                                'raw': {
                                    'type': 'keyword'
                                }
                            }
                        }
                    }
                }
            ),
            ('name', {'type': 'number'}, {'type': 'double'}),
            ('name', {'type': 'integer'}, {'type': 'long'}),
            (
                'name',
                {'type': 'string'},
                {
                    'type': 'text',
                    'fields': {'raw': {'type': 'keyword'}}
                }
            ),
            ('name', {'type': 'long_string'}, {'type': 'text'}),
        ]
    )
    def test_complex_types(
        self,
        field: str,
        field_def: Dict,
        expected_property: Optional[Dict]
    ):
        data = {
            'fields': {
                field: field_def,
            },
        }
        mapping = create_es6_mapping(data)
        assert field in mapping['properties']
        expected_property = expected_property or field_def
        assert mapping['properties'][field] == expected_property

    def test_sort(self):
        data = {
            'fields': {
                'name': {
                    'type': 'string'
                }
            },
            'sort': ['name']
        }

        mapping = create_es6_mapping(data)

        expected = {
            'dynamic': False,
            'properties': {
                'name': {
                    'type': 'text',
                    'fields': {
                        'raw': {
                            'type': 'keyword'
                        }
                    }
                }
            },
            'settings': {
                'analysis': {
                    'analyzer': {
                        'default': {
                            'char_filter': [
                                'compound',
                                'swedish_aa',
                                'swedish_ae',
                                'swedish_oe'
                            ],
                            'filter': [
                                'swedish_folding',
                                'lowercase'
                            ],
                            'tokenizer': 'standard'
                        },

                    },
                    'char_filter': {
                        'compound': {
                            'pattern': '-',
                            'replacement': '',
                            'type': 'pattern_replace'
                        },
                        'swedish_aa': {
                            'pattern': '[Ǻǻ]',
                            'replacement': 'å',
                            'type': 'pattern_replace'
                        },
                        'swedish_ae': {
                            'pattern': '[æÆǞǟ]',
                            'replacement': 'ä',
                            'type': 'pattern_replace'
                        },
                        'swedish_oe': {
                            'pattern': '[ØøŒœØ̈ø̈ȪȫŐőÕõṌṍṎṏȬȭǾǿǬǭŌōṒṓṐṑ]',
                            'replacement': 'ö',
                            'type': 'pattern_replace'
                        },
                    },
                    'filter': {
                        'swedish_folding': {
                            'type': 'icu_folding',
                            'unicodeSetFilter': '[^åäöÅÄÖ]'
                        },
                        'swedish_sort': {
                            'language': 'sv',
                            'type': 'icu_collation'
                        }
                    }
                }
            }
        }

        assert mapping == expected

    @pytest.mark.skip()
    def test_sort_large(self):
        data = {
            'fields': {
                'name': {
                    'type': 'string'
                }
            },
            'sort': ['name']
        }

        mapping = create_es6_mapping(data)

        expected = {
            'properties': {
                'name': {
                    'type': 'text',
                    'fields': {
                        'raw': {
                            'type': 'keyword'
                        }
                    }
                }
            },
            'settings': {
                'analysis': {
                    'analyzer': {
                        'default': {
                            'char_filter': [
                                'compound',
                                'swedish_aa',
                                'swedish_ae',
                                'swedish_oe'
                            ],
                            'filter': [
                                'swedish_folding',
                                'lowercase'
                            ],
                            'tokenizer': 'standard'
                        },

                    },
                    'char_filter': {
                        'clean_lastname': {
                            'pattern': '^(af|von) ',
                            'replacement': '',
                            'type': 'pattern_replace'
                        },
                        'compound': {
                            'pattern': '-',
                            'replacement': '',
                            'type': 'pattern_replace'
                        },
                        'init': {
                            'pattern': '^(.).*',
                            'replacement': '$1',
                            'type': 'pattern_replace'
                        },
                        'links': {
                            'pattern': '\\[([^\\]]*)\\](\\([^\\)]*\\))?',
                            'replacement': '$1',
                            'type': 'pattern_replace'
                        },
                        'pics': {
                            'pattern': '!\\[[^\\]]*\\](\\([^\\)]*\\))?',
                            'replacement': '',
                            'type': 'pattern_replace'
                        },
                        'remove_slash_filter': {
                            'pattern': '(.*)/(.+)/(.*)',
                            'replacement': '$1$2$3',
                            'type': 'pattern_replace'
                        },
                        'removespecials': {
                            'pattern': '[^A-Za-zÀÁÂÄÅÆÇÈÉÊËÌÍÎÏÑÒÓÔÖØÙÚÛÜàáâäåæçèéêëìíîïñòóôøöùúûü]',
                            'replacement': '',
                            'type': 'pattern_replace'
                        },
                        'slash_filter': {
                            'pattern': '(.*)/(.+)/(.*)',
                            'replacement': '$2',
                            'type': 'pattern_replace'
                        },
                        'swedish_aa': {
                            'pattern': '[Ǻǻ]',
                            'replacement': 'å',
                            'type': 'pattern_replace'
                        },
                        'swedish_ae': {
                            'pattern': '[æÆǞǟ]',
                            'replacement': 'ä',
                            'type': 'pattern_replace'
                        },
                        'swedish_oe': {
                            'pattern': '[ØøŒœØ̈ø̈ȪȫŐőÕõṌṍṎṏȬȭǾǿǬǭŌōṒṓṐṑ]',
                            'replacement': 'ö',
                            'type': 'pattern_replace'
                        },
                        'underscore': {
                            'pattern': '_*((~_)*)_*',
                            'replacement': '$1',
                            'type': 'pattern_replace'
                        }
                    },
                    'filter': {
                        'swedish_folding': {
                            'type': 'icu_folding',
                            'unicodeSetFilter': '[^åäöÅÄÖ]'
                        },
                        'swedish_sort': {
                            'language': 'sv',
                            'type': 'icu_collation'
                        }
                    }
                }
            }
        }

        assert mapping == expected




import logging
from os import stat
import traceback
import sys
import time
from typing import Any

try:
    from importlib.metadata import entry_points
except ImportError:
    # used if python < 3.8
    from importlib_metadata import entry_points  # type: ignore

from fastapi import FastAPI, Request, Response, status, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import JSONResponse
from fastapi.exception_handlers import http_exception_handler
import injector
from sqlalchemy.engine import Connection
from sqlalchemy.orm import Session
import logging
from asgi_correlation_id import CorrelationIdMiddleware
from asgi_correlation_id.context import correlation_id

from karp import main
from karp.foundation import errors as foundation_errors
from karp.foundation.value_objects import unique_id
from karp.auth import errors as auth_errors
from karp.lex.domain import errors as lex_errors
from karp.errors import ClientErrorCodes
from karp.main import modules, config
from karp.webapp.routes import router as api_router
from karp.webapp import tasks
from karp.webapp.contrib import MatomoMiddleware


querying_description = """
## Query DSL
### Query operators
- `contains|<field>|<string>` Find all entries where the field <field> contains <string>. More premissive than equals.

- `endswith|<field>|<string>` Find all entries where the field <field> ends with <string>

- `equals|<field>|<string>` Find all entries where <field> equals <string>. Stricter than contains

- `exists|<field>` Find all entries that has the field <field>.

- `freetext|<string>` Search in all fields for <string> and similar values.

- `freergxp|<regex.*>` Search in all fields for the regex <regex.*>.

- `gt|<field>|<value>` Find all entries where <field> is greater than <value>.

- `gte|<field>|<value>` Find all entries where <field> is greater than or equals <value>.

- `lt|<field>|<value>` Find all entries where <field> is less than <value>.

- `lte|<field>|<value>` Find all entries where <field> is less than or equals <value>.

- `missing|<field>` Search for all entries that doesn't have the field <field>.

- `regexp|<field>|<regex.*>` Find all entries where the field <field> matches the regex <regex.*>.

- `startswith|<field>|<string>` Find all entries where <field>starts with <string>.

### Logical Operators
The logical operators can be used both at top-level and lower-levels.

- `not(<expression1>||<expression2>||...)` Find all entries that doesn't match the expression <expression>.

- `and(<expression1>||<expression2>||...)` Find all entries that matches <expression1> AND <expression2>.

- `or(<expression1>||<expression2>||...)` Find all entries that matches <expression1> OR <expression2>.

### Regular expressions
Always matches complete tokens.

###  Examples
- `not(missing|pos)`
- `and(freergxp|str.*ng||regexp|pos|str.*ng)`
- `and(missing|pos||equals|wf||or|blomma|äpple`
- `and(equals|wf|sitta||not||equals|wf|satt`
"""


tags_metadata = [
    {
        "name": "Querying",
        "description": querying_description,
    },
    {"name": "Editing"},
    {"name": "Statistics"},
    {"name": "History"},
    {"name": "Resources"},
    {"name": "Health"},
]

logger = logging.getLogger(__name__)


def create_app() -> FastAPI:

    app_context = main.bootstrap_app()

    app = FastAPI(
        title=f"{config.PROJECT_NAME} API",
        redoc_url="/",
        version=config.VERSION,
        openapi_tags=tags_metadata,
    )

    app.state.app_context = app_context

    app.add_event_handler("startup", tasks.create_start_app_handler(app))
    app.add_event_handler("shutdown", tasks.create_stop_app_handler(app))

    main.install_auth_service(app_context.container, app_context.settings)

    app.add_middleware(
        CORSMiddleware,
        allow_origins=app_context.settings.get("web.cors.origins", ["*"]),
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    app.add_middleware(GZipMiddleware, minimum_size=1000)

    app.include_router(api_router)

    modules.load_modules("karp.webapp", app=app)

    from karp.errors import KarpError

    @app.exception_handler(KarpError)
    async def _karp_error_handler(request: Request, exc: KarpError):
        logger.exception(exc)
        return JSONResponse(
            status_code=exc.http_return_code,
            content={"error": exc.message, "errorCode": exc.code},
        )

    @app.exception_handler(foundation_errors.NotFoundError)
    async def _entity_not_found(request: Request, exc: foundation_errors.NotFoundError):
        return JSONResponse(
            status_code=404,
            content={
                "detail": str(exc),
                "type": str(type(exc)),
            },
        )

    @app.exception_handler(auth_errors.ResourceNotFound)
    async def _auth_entity_not_found(
        request: Request, exc: auth_errors.ResourceNotFound
    ):
        return JSONResponse(
            status_code=404,
            content={
                "detail": str(exc),
            },
        )

    @app.exception_handler(lex_errors.LexDomainError)
    def _lex_error_handler(request: Request, exc: lex_errors.LexDomainError):
        logger.exception(exc)
        return lex_exc2response(exc)

    @app.exception_handler(Exception)
    async def unhandled_exception_handler(request: Request, exc: Exception):
        return await http_exception_handler(
            request,
            HTTPException(
                status.HTTP_500_INTERNAL_SERVER_ERROR,
                detail="Internal server error",
                headers={
                    "X-Request-ID": correlation_id.get() or "",
                    "Access-Control-Expose-Headers": "X-Request-ID",
                },
            ),
        )

    @app.middleware("http")
    async def injector_middleware(request: Request, call_next):

        response: Response = JSONResponse(
            status_code=500, content={"detail": "Internal server error"}
        )
        try:
            request.state.connection = app_context.container.get(Connection)
            request.state.session = Session(bind=request.state.connection)
            request.state.container = app_context.container.create_child_injector(
                modules.request_configuration(
                    conn=request.state.connection,
                    session=request.state.session,
                )
            )

            response = await call_next(request)
        finally:
            connection = getattr(request.state, "connection", None)
            if connection:
                request.state.connection.close()

        return response

    @app.middleware("http")
    async def _logging_middleware(request: Request, call_next) -> Response:
        response: Response = JSONResponse(
            status_code=500, content={"detail": "Internal server error"}
        )
        start_time = time.time()
        try:
            response = await call_next(request)
        finally:
            process_time = time.time() - start_time
            logger.info(
                "processed a request",
                extra={
                    "status_code": response.status_code,
                    "process_time": process_time,
                },
            )
        return response

    if app_context.settings["tracking.matomo.url"]:
        app.add_middleware(
            MatomoMiddleware,
            idsite=app_context.settings["tracking.matomo.idsite"],
            matomo_url=app_context.settings["tracking.matomo.url"],
            access_token=app_context.settings["tracking.matomo.token"],
        )
    else:
        logger.warning(
            "Tracking to Matomo is not enabled, please set TRACKING_MATOMO_URL."
        )
    app.add_middleware(CorrelationIdMiddleware)

    return app


def lex_exc2response(exc: lex_errors.LexDomainError) -> JSONResponse:
    status_code = 503
    content: dict[str, Any] = {"message": "Internal server error"}
    if isinstance(exc, lex_errors.UpdateConflict):
        status_code = 400
        content = {
            "message": str(exc),
            "errorCode": ClientErrorCodes.VERSION_CONFLICT,
        }
    return JSONResponse(
        status_code=status_code,
        content=content,
    )



import pytest

import elasticsearch_dsl as es_dsl

from karp.search.domain.query_dsl import (
    KarpQueryV6Parser,
    KarpQueryV6ModelBuilderSemantics,
)
from karp.search_infrastructure.queries.es6_search_service import EsQueryBuilder


@pytest.fixture(scope="session")
def parser() -> KarpQueryV6Parser:
    return KarpQueryV6Parser(semantics=KarpQueryV6ModelBuilderSemantics())


@pytest.mark.parametrize(
    "q,expected",
    [
        ("exists|test", es_dsl.Q("exists", field="test")),
        ("freetext|hej", es_dsl.Q("multi_match", query="hej", fuzziness=1)),
        ("freetext|12", es_dsl.Q("multi_match", query=12)),
        (
            "freergxp|1i.*2",
            es_dsl.Q("query_string", query="/1i.*2/", default_field="*"),
        ),
        ("missing|test", es_dsl.Q("bool", must_not=es_dsl.Q("exists", field="test"))),
        ("startswith|pos|nn", es_dsl.Q("regexp", pos="nn.*")),
        ("endswith|pos|nn", es_dsl.Q("regexp", pos=".*nn")),
        ("contains|pos|nn", es_dsl.Q("regexp", pos=".*nn.*")),
        ("gt|val|lok", es_dsl.Q("range", val={"gt": "lok"})),
        ("gte|val|2", es_dsl.Q("range", val={"gte": 2})),
        ("lt|val|lok", es_dsl.Q("range", val={"lt": "lok"})),
        ("lte|val|lok", es_dsl.Q("range", val={"lte": "lok"})),
        ("equals|pos|vb", es_dsl.Q("match", pos={"query": "vb", "operator": "and"})),
        ("regexp|kjh|lk.*k", es_dsl.Q("regexp", kjh="lk.*k")),
        ("not(regexp|kjh|lk.*k)", ~es_dsl.Q("regexp", kjh="lk.*k")),
        (
            "and(regexp|baseform|g[oe]t||equals|pos|nn)",
            es_dsl.Q("regexp", baseform="g[oe]t")
            & es_dsl.Q("match", pos={"query": "nn", "operator": "and"}),
        ),
        (
            "and(regexp|baseform|g[oe]t||equals|pos|nn||regexp|pos|n.*)",
            es_dsl.Q("regexp", baseform="g[oe]t")
            & es_dsl.Q("match", pos={"query": "nn", "operator": "and"})
            & es_dsl.Q("regexp", pos="n.*"),
        ),
        (
            "or(regexp|baseform|g[oe]t||equals|pos|nn)",
            es_dsl.Q("regexp", baseform="g[oe]t")
            | es_dsl.Q("match", pos={"query": "nn", "operator": "and"}),
        ),
        (
            "or(regexp|baseform|g[oe]t||equals|pos|nn||regexp|pos|n.*)",
            es_dsl.Q("regexp", baseform="g[oe]t")
            | es_dsl.Q("match", pos={"query": "nn", "operator": "and"})
            | es_dsl.Q("regexp", pos="n.*"),
        ),
        # (
        #     "regexp|or(pos||form)|k.+l",
        #     es_dsl.Q("regexp", pos="k.+l") | es_dsl.Q("regexp", form="k.+l")
        # ),
        # (
        #     "count(baseform)|gt|1",
        #     None,
        # )
    ],
)
def test_es_query(parser, q, expected):
    model = parser.parse(q)

    query = EsQueryBuilder().walk(model)

    assert query == expected


@pytest.mark.parametrize(
    "q,expected",
    [
        (
            "and(exists|pos||exists|vb)",
            es_dsl.Q(
                "bool",
                must=[
                    es_dsl.Q("exists", field="pos"),
                    es_dsl.Q("exists", field="vb"),
                ],
            ),
        ),
        (
            "or(exists|pos||exists|vb)",
            es_dsl.Q(
                "bool",
                should=[
                    es_dsl.Q("exists", field="pos"),
                    es_dsl.Q("exists", field="vb"),
                ],
            ),
        ),
        (
            "and(not(exists|pos)||not(exists|vb))",
            es_dsl.Q(
                "bool",
                must_not=[
                    es_dsl.Q("exists", field="pos"),
                    es_dsl.Q("exists", field="vb"),
                ],
            ),
        ),
        (
            "not(equals|ordklass|substantiv||equals|ordklass|verb)",
            es_dsl.Q(
                "bool",
                must_not=[
                    es_dsl.Q(
                        "match", ordklass={"query": "substantiv", "operator": "and"}
                    ),
                    es_dsl.Q("match", ordklass={"query": "verb", "operator": "and"}),
                ],
            ),
        ),
        #     (
        #         "exists|or(pos||vb)",
        #         es_dsl
        #         )
    ],
)
def test_combined_es_query(parser, q, expected):
    query = EsQueryBuilder().walk(parser.parse(q))

    assert query == expected
